from PymoNNto.NetworkBehaviour.Input.Pattern_Basics import *
import random
import numpy as np
from collections import Counter
import pypianoroll as piano
import os
import re
from sklearn.feature_extraction.text import CountVectorizer
from scipy.stats import entropy
import csv
from scipy.stats import entropy
from itertools import groupby, count


#{' PERIOD': '.', ' COMMA': '.', ' QUESTION': '.'} #, 'MNAME':'Name', 'FNAME':'Name'
#CHILDES_activator = TextActivator_New(filename='CHILDES.txt', replace_dict={' PERIOD': '.', ' COMMA': '.', ' QUESTION': '.'})
#SPD_activator = TextActivator_New(filename='SPD.txt', replace_dict={'?': '.', ':': '.', ',': '.', '.': '. ', '  ': ' ', '-': ''})

class DrumBeatActivator(PatternGroup):

    def initialize(self): #
        self.last_iteration=-1
        self.nextone=None
        self.is_drum = True
        self.type = 'poly'

        self.offtoken = self.kwargs.get('offtoken', False)
        self.ontoken = self.kwargs.get('ontoken', False)
        self.include_inverse_alphabet = self.kwargs.get('include_inverse_alphabet', False)
        self.max_iterations = self.kwargs.get('max_iterations', None)
        self.filter_silence = self.kwargs.get('filter_silence', False) # false if no filter should be applied, integer for maximum allowed silent time steps


        self.beat_resolution = self.kwargs.get('beat_resolution', 4)

        self.current_index = -1
        self.seen = [] # part of corpus that has already been seen
        self.times_seen = 0 # times we have iterated through whole corpus

        self.which_tracks = self.kwargs.get('which_tracks', 'all') # read in all tracks in corpus, otherwise a list of indices
        self.distance_similar_tacks = self.kwargs.get('THR_similar_tracks', None) # define threshold for euclidean distance in 2D space to select similar tracks        

        if self.distance_similar_tacks is not None:
            assert len(self.which_tracks) == 1 # selecting similar tracks only implemented if one example track is given as input 
            
        self.path_to_music = self.kwargs.get('path_to_music', os.path.join(os.path.dirname(__file__),'..', '..','..','Data/midis/beats/'))
        self.alphabet = self.kwargs.get('instruments', [36,38,41,42,45,46,48,49,51,60,61,62,64,69]) # list of MIDI percussion indices

        self.instruments = self.get_instruments(self.path_to_music, self.alphabet) # dictionary of instruments
        self.corpus_blocks = self.get_beatmatrix(self.path_to_music, 'dataset.tsv', self.alphabet, self.which_tracks, self.max_iterations) # sequence of vectors with shape (128)

        if self.offtoken:
            self.alphabet.append(0) # STOP TOKEN
            self.instruments[0] = 'STOP'
        
        if self.ontoken:
            self.alphabet.append(1) # START TOKEN
            self.instruments[1] = 'START'

        self.corpus = self.create_corpus_from_tracks(self.corpus_blocks, self.offtoken, self.ontoken)
        #self.corpus_blocks.reshape((int(self.corpus_blocks.shape[0]*self.corpus_blocks.shape[1]), self.corpus_blocks.shape[2]))

        self.A = len(self.alphabet)
        if self.include_inverse_alphabet:
            self.A *= 2

    def create_corpus_from_tracks(self, tracks, offtoken, ontoken):
        corpus = tracks
        if offtoken:
            # we augment the corpus by a stop token 
            stop_row = np.zeros((tracks.shape[0], tracks.shape[1], 1))
            stop_row[:,-1,0] = 1
            corpus = np.concatenate((corpus, stop_row), axis=2)
        if ontoken: 
            # we augment the corpus by a start token
            start_row = np.zeros((tracks.shape[0], tracks.shape[1], 1))
            start_row[:,0,0] = 1
            corpus = np.concatenate((corpus, start_row), axis=2)

        corpus = corpus.reshape((int(corpus.shape[0]*corpus.shape[1]), corpus.shape[2])) # flatten beatmatrix

        return corpus

    def get_instruments(self, path, alphabet):
        all_instruments = {}
        with open(os.path.join(path, 'instruments.txt')) as f:
            for line in f:
                all_instruments[int(line[:2])] = line[3:-1]

        dict_instruments = {}
        for i in alphabet:
            dict_instruments[i] = all_instruments[i]

        return dict_instruments

    def get_beatmatrix(self, path, file, instruments, which_tracks='all', max_length=None):
        # creates polyphonic pianoroll of length max iterations
        dataset = []
        coordinates = []

        if which_tracks != 'all':

            # read tracks from list which_tracks
            with open(os.path.join(path, file)) as tsvfile:
                reader = csv.reader(tsvfile, delimiter='\t')
                dataset = [np.array(row[0].split(','),int) for idx, row in enumerate(reader) if idx in which_tracks]
            with open(os.path.join(path, file)) as tsvfile:
                reader = csv.reader(tsvfile, delimiter='\t')
                coordinates = [np.array(row[-1].split(','), float) for idx, row in enumerate(reader) if idx in which_tracks]

                # now choose similar tracks (with euclidian distance <= THR in 2D VAE projection)
                if self.distance_similar_tacks is not None: # select all tracks that are similar to the track in  self_which_tracks
                    with open(os.path.join(path, file)) as tsvfile:
                        reader = csv.reader(tsvfile, delimiter='\t')
                        for row in reader:
                            check = np.array(row[-1].split(','), float)
                            if np.linalg.norm(coordinates[0]-check) < self.distance_similar_tacks and np.linalg.norm(coordinates[0]-check)!=0:
                                coordinates.append(check)
                                dataset.append(np.array(row[0].split(','), int))
        else: # load all tracks 
            with open(os.path.join(path, file)) as tsvfile:
                reader = csv.reader(tsvfile, delimiter='\t')
                for row in reader:
                    dataset.append(np.array(row[0].split(','),int))
                    coordinates.append(np.array(row[-1].split(','), float))

        dataset = np.array(dataset)
        coordinates = np.array(coordinates)

        #if which_tracks is not 'all':
        #    dataset = dataset[which_tracks,:]

        if max_length is not None: 
            if max_length < dataset.shape[0]*dataset.shape[1]:
                n_rows = int(max_length/dataset.shape[1]) # we only take full tracks 
                dataset = dataset[:n_rows, :]

        if self.filter_silence: 
            # remove tracks that have longer silence than specified by self.filter_silence
            dataset = self.filter_tracks_by_max_silence(dataset, self.filter_silence)

        print(len(dataset), 'tracks')
        

        # now we create the matrix from the bit strings (each row in dataset is a matrix)
        beat_matrices = np.zeros((dataset.shape[0], dataset.shape[1], 14), bool)

        for i, row in enumerate(dataset):
            for j in range(len(row)):
                binary_string = np.binary_repr(dataset[i,j], width=len(instruments))
                # now populate the matrix
                for k, letter in enumerate(binary_string):
                    if int(letter):
                        beat_matrices[i, j, -k] = True

        return beat_matrices
        
    def get_instrument_input_statistics_list(self):
        stats = [len(np.nonzero(self.corpus[:,i])[0]) for i in range(len(self.alphabet))]
        stats = np.array(stats)
        if self.include_inverse_alphabet: # append the frequencies with which symbols do not occur
            inverse_symbols = np.repeat(len(self.corpus), len(stats)) - stats
            stats = np.concatenate((stats, inverse_symbols))        
        return stats


    #def get_alphabet_length(self):
    #    return len(self.alphabet)

    def generate_weights(self, neurons):
        if hasattr(neurons, 'input_density'):
            density = neurons.input_density
        else:
            density = self.kwargs.get('input_density', 1/60)#int(output_size / 60)

        output_size = neurons.size
        #output_size = self.kwargs.get('output_size', 600)
        # self.kwargs.get('activation_size', int(output_size / 60))

        A = self.A

        self.activation_size = int(output_size * density)

        neurons.Input_Weights = np.zeros((output_size, A))
        available = set(range(output_size))
        for a in range(A):
            temp = random.sample(available, self.activation_size)
            neurons.Input_Weights[temp, a] = 1
            available = set([n for n in available if n not in temp])
            assert len(available) > 0, 'Input alphabet too big for non-overlapping neurons'

        neurons.Input_Mask = (np.sum(neurons.Input_Weights, axis=1) > 0)
        neurons.mean_network_input_activity=self.activation_size/output_size

    def get_pattern(self, neurons): 
        if self.active:
            n_hot = self.get_next(next=neurons.iteration != self.last_iteration)
            self.last_iteration=neurons.iteration
            self.current_pattern_index = n_hot #for pattern activator   --> how does this work if my pattern is more than one symbol at a time?
            return self.get_activation(n_hot, neurons) #self.W[:, i]
        else:
            return None

    def initialize_neuron_group(self, neurons):
        #if not hasattr(neurons, 'Input_Weights'):
        self.generate_weights(neurons)

    
    def albphabet_index_to_note_symbol(self, index):
        """
        Convert an alphabet index (from the network) to the corresponding note string

        Arguments:
        index -- int, alphabet index

        Returns:
        symbol -- string corresponding to note
        """

        # first get the MIDI index
        midi = self.alphabet[index]
        return self.instruments[midi]

    
    def midi_index_to_alphabet_index(self, midi):
        '''
        looks for the MIDI index in the alphabet of the SORN and returns its index in the alphabet

        Arguments:
        midi -- int, a MIDI index

        Returns:
        index -- int, an index to index the networks alphabet
        '''
        return self.alphabet.index(midi)

    def convert_sequence_to_pianoroll(self, sequence):
        'conerts a sequence of MIDI indices to a pianoroll'
        pianoroll = np.zeros((len(sequence), 128)).astype(np.bool_)
        alphabet = self.alphabet
        if self.ontoken:
            alphabet = np.setdiff1d(alphabet, 1)
        if self.offtoken:
            alphabet = np.setdiff1d(alphabet, 0)
        pianoroll[:, alphabet] = sequence
        return pianoroll

    def set_nextone(self, vec):
        'sets next vector'
        self.nextone = vec


    def get_next(self, next=False):
        if next:
            self.current_index += 1

        # in case the sequence ends, restart it
        if self.current_index == len(self.corpus):
            self.current_index = 0
            self.times_seen+=1

        if self.nextone is not None:#outside activation
            self.corpus[self.current_index,:] = self.nextone

            self.nextone=None
        
        if next:
            self.seen.append(self.corpus[self.current_index,:].astype(bool))

            # corpus is already sequence of n-hot arrays, we do not need to convert it

        next_vector = self.corpus[self.current_index,:].astype(bool)

        if self.include_inverse_alphabet: 
            next_vector = np.concatenate((next_vector, np.invert(next_vector)))

        return next_vector
        #return self.corpus[self.current_index,:].astype(bool)

    def get_activation(self, n_hot, neurons):
        if not np.any(n_hot):
            return np.zeros(neurons.Input_Weights.shape[0]).astype(bool) # return a vector of zeros
        else:
            vector = np.zeros(neurons.Input_Weights.shape[0])
            indices = np.where(n_hot == 1)[0]
            for i in indices:
                vector += neurons.Input_Weights[:,i]
            return vector.astype(bool)
            #return neurons.Input_Weights[:,n_hot]
            #return neurons.Input_Weights * mask

    def get_music_score(self, spont_output, pianoroll, seen=None):
        result = {}

        output = spont_output[:-1]
        output=output.split(' ')

        print(set(np.nonzero(np.array(self.seen)[:len(self.seen)-len(output),:])[1]))

        if seen is None: # no network output sequence was given
            seen = self.seen[:len(self.seen)-len(output)] # the last n time steps it gets fed in the previous spontaneous output: that's not comparable between experiments

        seen_string = ''
        for i in range(len(seen)-len(output)): # the last n time steps it gets fed in the previous spontaneous output: that's not comparable between experiments
            if np.any(seen[i]): # vector is not all zeros
                indices = np.nonzero(seen[i])[0]
                if len(indices)==1: # predict only one note at this time step
                    note = self.albphabet_index_to_note_symbol(indices[0])
                    seen_string += note+' '
                else:
                    for j in indices:
                        note = self.albphabet_index_to_note_symbol(j)
                        seen_string += note
                    seen_string += ' '
            else:
                seen_string += '_ ' # silence

        seen_string = seen_string.split(' ')
        seen_string = seen_string[:-1]

        #result['seen_string']=seen_string
        result['times_seen'] = self.times_seen
        result['spont_output'] = spont_output
        result['output_beatmatrix'] = pianoroll

        # compute per track entropy and mean length
        if self.ontoken or self.offtoken:
            tracks = self.create_beat_matrix_from_results(pianoroll, self.offtoken, self.ontoken)
            result['track_length'] = self.compute_per_track_length(tracks)
            result['track_entropy'] = self.compute_per_track_entropy(tracks)
            print('mean per track entropy:', np.mean(result['track_entropy']), 'std:', np.std(result['track_entropy']))
            print('mean track length:', np.mean(result['track_length']), 'std:', np.std(result['track_length']))

        #result['seen_beatmatrix'] = seen
        result['alphabet'] = self.alphabet
        result['total_score']=0

        # compare frequencies of notes 

        # make a list of all possible notes 
        indices_non_zero = np.nonzero(pianoroll)
        number_note_played = {}
        number_note_seen = {}

        indices_non_zero_seen = np.nonzero(np.array(seen))

        for note in set(indices_non_zero_seen[1]): # can't take alphabet bc perhaps it has not seen all of corpus
            number_note_played[self.albphabet_index_to_note_symbol(note)] = len(indices_non_zero[0][np.where(indices_non_zero[1]==note)])
            number_note_seen[self.albphabet_index_to_note_symbol(note)] = len(indices_non_zero_seen[0][np.where(indices_non_zero_seen[1]==note)])

        number_note_seen['_'] = len(set(range(len(seen))) - set(indices_non_zero_seen[0]))
        number_note_played['_'] = len(set(range(len(pianoroll))) - set(indices_non_zero[0]))

        sorted_notes = [x[0] for x in sorted(number_note_seen.items(), key=lambda kv: kv[1], reverse=True)]
        freq_input = np.array([number_note_seen[x] for x in sorted_notes])
        freq_input = freq_input/freq_input.sum()

        result['input_frequencies'] = inp = dict(zip(sorted_notes, freq_input))
        result['sorted_symbols'] = sorted_notes

        freq_output = np.array([number_note_played[x] for x in sorted_notes])
        freq_output = freq_output/freq_output.sum()

        result['output_frequencies'] = out = dict(zip(sorted_notes, freq_output))

        result['diff_frequencies'] = diff = {x: abs(inp[x] - out[x]) for x in inp if x in out}

        result['sum_differences'] = sum_diff = sum(diff.values())

        result['total_score'] -= sum_diff

        # Now we look at the frequencies of chords

        times_seen_chords = {}
        notes_in_chords = {}
        for ts in range(len(seen_string)):
            if len(np.where(seen[ts])[0])>1: # only a chord if more than one note is played at the same time
                chord = seen_string[ts]
                if chord in times_seen_chords:
                    times_seen_chords[chord] +=1
                else:
                    notes_in_chords[chord] = np.where(seen[ts])[0] # how many notes are in this chord?
                    times_seen_chords[chord] = 1
          
        times_output_chords = {}

        for ts in range(len(pianoroll)):
            if len(np.where(pianoroll[ts])[0])>1:
                chord = output[ts]
                if chord in times_output_chords:
                    times_output_chords[chord] += 1
                else:
                    times_output_chords[chord] = 1
                if chord not in notes_in_chords:
                    notes_in_chords[chord] = np.where(pianoroll[ts][0])

        sorted_chords = [x[0] for x in sorted(times_seen_chords.items(), key=lambda kv: kv[1], reverse=True)]
        freq_input = np.array([times_seen_chords[x] for x in sorted_chords])
        freq_input = freq_input/freq_input.sum()

        result['input_frequencies_chords'] = inp = dict(zip(sorted_chords, freq_input))
        result['sorted_chords'] = sorted_chords

        freq_output = np.zeros(len(sorted_chords))
        for i, chord in enumerate(sorted_chords):
            if chord in times_output_chords:
                freq_output[i] = times_output_chords[chord]

        #freq_output = np.array([times_output_chords[x] for x in sorted_chords])
        freq_output = freq_output/freq_output.sum()

        result['output_frequencies_chords'] = out = dict(zip(sorted_chords, freq_output))
        result['midi_indices_per_chord'] = notes_in_chords
        result['n_notes_per_chord'] = len(notes_in_chords)

        # differences in frequencies
        # punish less if it gets frequencies of chords (compared to notes overall)
        result['diff_frequencies_chords'] = diff = {x: abs(inp[x] - out[x]) for x in inp if x in out}

        sum_diff = 0
        
        for key in diff: # punish less if it gets frequencies of higher n_grams wrong
            sum_diff += diff[key]/len(notes_in_chords[key]) # weighted by number of notes in a chord
        result['sum_differences_chords'] = sum_diff

        result['total_score'] -= sum_diff 

        #output frequencies for all output chords (also ones that do not appear in input)
        sorted_chords = [x[0] for x in sorted(times_output_chords.items(), key=lambda kv: kv[1], reverse=True)]
        freq_output = np.array([times_output_chords[x] for x in sorted_chords])
        freq_output = freq_output/freq_output.sum()

        result['sorted_chords_output_all'] = sorted_notes
        result['output_frequencies_chords_all'] = freq_output

        return result

    def get_entropy_of_track(self, pianoroll):
        alphabet = self.alphabet()
        indices_non_zero = pianoroll.nonzero()
        distribution = np.zeros(len(alphabet))
        for i in range(len(alphabet)):
            distribution[i] = sum(indices_non_zero[1] == alphabet[i])
        # add the silent time steps
        non_zero_indices = set(range(indices_non_zero[0]))
        all_indices = set(range(len(pianoroll)))
        silent_indices = all_indices - non_zero_indices
        distribution=np.append(distribution, len(silent_indices))
    
        return entropy(distribution)

    def create_beat_matrix_from_results(self, beat_matrix, offtoken=True, ontoken=False):
        #TODO: need to check that indexing is safe if start/stop tokens occur at first/last time step of output

        if offtoken and ontoken:
            # last row stands for start token, row before for stop token
            indices_start = np.nonzero(beat_matrix[:,-1])[0]
            indices_stop = np.nonzero(beat_matrix[:,-2])[0]

        elif offtoken and not ontoken: # we only have offtokens, but we start at first generated time step
            indices_stop = np.nonzero(beat_matrix[:,-1])[0]
            indices_start = indices_stop+1

        elif ontoken and not offtoken:
            indices_start = np.nonzero(beat_matrix[:,-1])[0]
            indices_stop = indices_start-1

        indices_start = indices_start[:-1] # remove last start token, it does not have a corresponding stop token
        indices_stop = indices_stop[1:] # remove first stop token, it does not have a corresponding start token

        tracks = []
        if len(indices_start) >0 and len(indices_stop)>0:
            assert [indices_stop[i] > indices_start[i] for i in range(len(indices_start))]

            for i in range(len(indices_start)):
                tracks.append(beat_matrix[indices_start[i]:indices_stop[i]+1,:14])
        
        else:# return whole matrix, we just have one track, it never predicted start/stop
            tracks.append(beat_matrix[:,:14])
        
        return np.array(tracks) 


    def get_percussion_track_entropy(self, track, len_alphabet=14, include_silence=False):
        distribution = np.zeros(len_alphabet)
        indices_non_zero = track.nonzero()
        for i in range(len_alphabet):
            distribution[i] = sum(indices_non_zero[1] == i)
        if include_silence:
        # add the silent time steps (when no note is played)
            non_zero_indices = set(indices_non_zero[0])
            all_indices = set(range(len(track)))
            silent_indices = all_indices - non_zero_indices
            distribution = np.append(distribution, len(silent_indices))

        return entropy(distribution)

    def compute_per_track_entropy(self, tracks, include_silence=False):
            '''
            return a list of per track entropies
            expects a beat matrix of shape (n_tracks, len_alphabet)
            '''
            entropy_generated_tracks = []

            if len(tracks)==1:
                entropy_generated_tracks = self.get_percussion_track_entropy(tracks, tracks.shape[1], include_silence = include_silence)

            else:
                for i in range(len(tracks)):
                    entropy_generated_tracks.append(self.get_percussion_track_entropy(tracks[i], tracks[i].shape[1], include_silence=include_silence))

            return entropy_generated_tracks

    def compute_per_track_length(self, tracks):
        '''
        returns a list of track lengths
        expects a beat matrix of shape (n_tracks, len_alphabet)
        '''
        return [len(i) for i in tracks]

    def find_length_longest_silence(self, track):
        '''
        given a beat_matrix (shape ts, instruments),
        returns the length of the longest silence,
        i.e. consecutive time steps where no instrument is played.
        '''
        silent_ts = [t for t in range(len(track)) if not np.any(track[t])]
        if silent_ts:
            c = count()
            val = max((list(g) for _, g in groupby(silent_ts, lambda x: x-next(c))), key=len)
            return len(val)
        else:
            return 0

    def filter_tracks_by_max_silence(self, corpus, max_silence=2):
        '''
        filters a corpus (of beatmatrices) by removing tracks that have
        more time steps of silence than specified by max_silence.
        '''
        filtered_corpus = [track for track in corpus if self.find_length_longest_silence(track)<=max_silence]
        return np.array(filtered_corpus)


from PymoNNto.Exploration.Network_UI.TabBase import *

def pol2cart(theta, rho):
    x = rho * np.cos(theta)
    y = rho * np.sin(theta)
    return x, y

def cart2pol(x, y):
    theta = np.arctan2(y, x)
    rho = np.hypot(x, y)
    return theta, rho

class DrawItem2(pg.GraphicsObject):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.picture = QtGui.QPicture()

    #base_char_masks_dict

    def draw_rings(self, painter, alphabet, num_rings):

        attractor_rads = []
        for i in range(num_rings):
            c=int(200+i*5.5)
            painter.setPen(pg.mkPen(color=(c, c, c, 255)))
            r = 100*np.power(0.9, i)
            attractor_rads.append(r)

            painter.setBrush(pg.mkBrush(color=(c, c, c, 0)))
            painter.drawEllipse(QtCore.QPointF(0, 0), r, r)

            for d, char in enumerate(alphabet):
                x, y = pol2cart(d / len(alphabet) * 2 * np.pi, r)
                painter.setBrush(pg.mkBrush(color=(c, c, c, 255)))
                painter.drawEllipse(QtCore.QPointF(x, y), 1, 1)
        return  attractor_rads

    def compute_and_apply_attraction(self, group, movement_speed_fac, anti_gravity_exp, random_movement_fac, weight_exp, attractor_rad_fac, attractor_rads):

            #group.buffer_posx += (group.get_random_neuron_vec() - 0.5)*0.1
            #group.buffer_posy += (group.get_random_neuron_vec() - 0.5)*0.1

            group.add_x = (group.get_random_neuron_vec() - 0.5) * random_movement_fac# + (-group.buffer_posx)*p2
            group.add_y = (group.get_random_neuron_vec() - 0.5) * random_movement_fac# + (-group.buffer_posy)*p2#np.sin(group.buffer_posy/lb*2*np.pi)*10*p2

            for sg in group.afferent_synapses['GLU']:
                theta_src, rho_src = cart2pol(sg.src.buffer_posx, sg.src.buffer_posy)

                if hasattr(group, 'timescale'):
                    step_factor = np.power(0.9, group.timescale)
                else:
                    step_factor = 0.9

                attractorx, attractory = pol2cart(theta_src, rho_src * step_factor)

                theta_dst, rho_dst = cart2pol(sg.dst.buffer_posx, sg.dst.buffer_posy)
                rd = np.array(attractor_rads)[:, None]-rho_dst[None, :]

                #get closest attractor
                if attractor_rad_fac > 0:
                    closest = np.array([rd[indx, i] for i, indx in enumerate(np.argmin(np.abs(rd), axis=0))])
                    #move_to_rad = np.choose(np.argmax(np.abs(rd), axis=1), rd)
                    attractor_rad_x, attractor_rad_y = pol2cart(theta_dst, closest)

                    f = np.abs(attractor_rad_x) + np.abs(attractor_rad_y)
                    attractor_rad_x[f > 0] /= f[f > 0]/len(attractor_rad_x)
                    attractor_rad_y[f > 0] /= f[f > 0]/len(attractor_rad_y)

                xdiff_mat = attractorx[:, None]-sg.dst.buffer_posx[None, :]
                ydiff_mat = attractory[:, None]-sg.dst.buffer_posy[None, :]

                f = np.abs(xdiff_mat)+np.abs(ydiff_mat)
                xdiff_mat[f > 0] /= f[f > 0]/len(xdiff_mat)
                ydiff_mat[f > 0] /= f[f > 0]/len(ydiff_mat)

                xdiff_mat[(f < 3) * (f > 0)] *= 1/3*(f[(f < 3) * (f > 0)])
                ydiff_mat[(f < 3) * (f > 0)] *= 1/3*(f[(f < 3) * (f > 0)])

                #d=np.sqrt(xdiff_mat*xdiff_mat+ydiff_mat*ydiff_mat)

                #xdiff_mat[d < 0.1] *= -10
                #ydiff_mat[d < 0.1] *= -10

                #print(rho)

                rho_src = np.power(rho_src, anti_gravity_exp)
                rho_src = rho_src/np.sum(rho_src)*len(rho_src)

                xdiff_mat *= rho_src[:, None]
                ydiff_mat *= rho_src[:, None]

                #xdiff_mat=xdiff_mat*xdiff_mat*xdiff_mat
                #ydiff_mat=ydiff_mat*ydiff_mat*ydiff_mat

                weights = np.power(sg.W, weight_exp)
                weights = weights/np.sum(weights)*len(weights)

                ax = 0
                sg.dst.add_x += np.sum(weights.T*xdiff_mat, axis=ax)/xdiff_mat.shape[ax]
                sg.dst.add_y += np.sum(weights.T*ydiff_mat, axis=ax)/ydiff_mat.shape[ax]

                if attractor_rad_fac>0:
                    sg.dst.add_x += attractor_rad_x*attractor_rad_fac*0.001
                    sg.dst.add_y += attractor_rad_y*attractor_rad_fac*0.001

            group.buffer_posx = np.clip(group.buffer_posx + group.add_x * movement_speed_fac, -100, +100)
            group.buffer_posy = np.clip(group.buffer_posy + group.add_y * movement_speed_fac, -100, +100)




    def draw_neurons(self, painter, group, neuron_size=1):
        painter.setPen(pg.mkPen(color=group.color))
        painter.setBrush(pg.mkBrush(color=group.color))
        mask=group.output>0
        inv_mask=np.invert(mask)
        for x, y in zip(group.buffer_posx[inv_mask], group.buffer_posy[inv_mask]):
            painter.drawEllipse(QtCore.QPointF(x, y), neuron_size, neuron_size)

        painter.setPen(pg.mkPen(color=(0,255,0,150)))
        painter.setBrush(pg.mkBrush(color=(0,255,0,150)))
        for x, y in zip(group.buffer_posx[mask], group.buffer_posy[mask]):
            painter.drawEllipse(QtCore.QPointF(x, y), neuron_size, neuron_size)

    def draw_weights(self, painter, group, selected_neuron_id):
        painter.setPen(pg.mkPen(color=(255, 0, 0, 255)))
        painter.setBrush(pg.mkBrush(color=(0, 0, 0, 0)))

        for sg in group.afferent_synapses['GLU']:
            temp = np.where(sg.dst.id == selected_neuron_id)[0]
            if len(temp) > 0:
                indx = temp[0]
                x = sg.dst.buffer_posx[indx]
                y = sg.dst.buffer_posy[indx]
                weights = sg.W[indx, :]
                max_w = np.max(weights)
                # painter.drawEllipse(QtCore.QPointF(x, y), 0.1, 0.1)
                for i, w in enumerate(weights):
                    if w > 0.0:
                        xs = sg.src.buffer_posx[i]
                        ys = sg.src.buffer_posy[i]
                        if xs != x and ys != y:
                            cv = 255.0 / max_w * w
                            cv = (cv * cv) / (255)
                            painter.setPen(pg.mkPen(color=(255, 0, 0, cv)))
                            painter.drawLine(QtCore.QPointF(x, y), QtCore.QPointF(xs, ys))  #


    def fixate_points(self, group, alphabet, input_mat):
        for i, char in enumerate(alphabet):
            bm=input_mat[:, i]
            x, y = pol2cart(i / len(alphabet) * 2 * np.pi, 100)
            group.buffer_posx[bm > 0.1] = x
            group.buffer_posy[bm > 0.1] = y


    def draw_chars(self, painter, alphabet, statistics):
        for i, char in enumerate(alphabet):
            x, y = pol2cart(i / len(alphabet) * 2 * np.pi, 110)

            val = float(statistics[i]) / float(np.max(statistics)) * 20.0

            qf=QFont('Arial')
            qf.setPointSizeF(2+val/4.0)
            #qf.setStretch(2)
            painter.setFont(qf)
            if char==' ':
                c='_'
            else:
                c=char
            painter.drawText(x,y,c)


    def update_pic(self, groups, alphabet, p0, p1, p2, p3, p4, p5, nui, statistics, show_weights):

        self.picture = QtGui.QPicture()
        painter = QtGui.QPainter(self.picture)

        attractor_rads=self.draw_rings(painter,alphabet,10)

        self.groups=groups

        #for group in groups:
        #    self.initialize_neuron_positons(group)

        for group in nui.network.NeuronGroups:
            self.compute_and_apply_attraction(group, p1, p2, p3, p4, p5, attractor_rads)

        painter.scale(1, -1)

        for group in groups:
            self.draw_neurons(painter, group, p0)

        #painter.setBrush(pg.mkBrush(color=(0,255,0,255)))
        #for x, y in zip(group.buffer_posx+attractor_rad_x, group.buffer_posy+attractor_rad_y):
        #    painter.drawEllipse(QtCore.QPointF(x, y), 1, 1)

        if show_weights:
            for group in groups:
                if group.tags[0] == nui.neuron_select_group:
                    self.draw_weights(painter, group, nui.neuron_select_id)


        painter.setPen(pg.mkPen(color=(0, 0, 0, 255), width=10))
        painter.setBrush(pg.mkBrush(color=(0, 0, 0, 0)))

        for group in nui.network.NeuronGroups:
            if hasattr(group, 'Input_Weights'):
                self.fixate_points(group, alphabet, group.Input_Weights)

        self.draw_chars(painter, alphabet, statistics)

        painter.end()
        self.prepareGeometryChange()
        self.informViewBoundsChanged()
        #self.update()

    def paint(self, painter, option, widget=None):
        painter.drawPicture(0, 0, self.picture)

    def boundingRect(self):
        br = self.picture.boundingRect()
        br.setTop(max(br.top()-10, br.left()-10))
        br.setLeft(min(br.top()-10, br.left()-10))
        br.setRight(max(br.right()+10, br.bottom()+10))
        br.setBottom(max(br.right()+10, br.bottom()+10))
        return QtCore.QRectF(br)

class sun_gravity_plot_tab(TabBase):

    def __init__(self, title='Sun Gravity Plot'):
        super().__init__(title)

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):
        if Network_UI.network['grammar_act', 0] is not None or Network_UI.network['drum_act', 0] is not None or Network_UI.network['music_act', 0] is not None:
            self.sun_gravity_plot_tab = Network_UI.Next_Tab(self.title)

            #alphabet = Network_UI.network['grammar_act'][0].alphabet
            #a_list = [alphabet[i] for i in range(len(alphabet))]
            #a_list[0] = '_'
            #ydict = dict(enumerate(a_list))

            #win = pg.GraphicsWindow()
            #stringaxis = pg.AxisItem(orientation='left')
            #stringaxis.setTicks([ydict.items()])

            for group in  Network_UI.network.NeuronGroups:
                group.buffer_posx, group.buffer_posy = pol2cart(group.get_random_neuron_vec() * 2 * np.pi, group.get_random_neuron_vec() * 100)

            def c(event):
                for group in event.currentItem.groups:
                    dx = group.buffer_posx-event.pos().x()
                    dy = group.buffer_posy-event.pos().y()*-1
                    d = np.sqrt(dx*dx+dy*dy)

                    indices = np.where(d<1)[0]

                    if len(indices) > 0:
                        Network_UI.neuron_select_group = group.tags[0]
                        Network_UI.neuron_select_id = indices[0]

            msg = 'Each particle is a neuron of the selected Group.\r\n' \
                  'The primary input neurons are fixed to the outer ring.\r\n' \
                  'The attraction between the neuron-particles is defined by synaptic weights.\r\n' \
                  'To avoid clustering in the center, the attraction increases exponetially with greater distance from the center.\r\n' \
                  'The point of attraction is not in the center of each particle,\r\n' \
                  'but the particlses position moved one step furter to the center, so one ring indicates one iteration step.'

            self.plot = Network_UI.Add_plot(x_label='buffer steps', tooltip_message=msg)#axisItems={'left': stringaxis}
            self.draw_item = DrawItem2()
            self.plot.addItem(self.draw_item)
            self.draw_item.mouseClickEvent = c

            if Network_UI.network['grammar_act', 0] is not None: 
                source = Network_UI.network['grammar_act'][0]
                self.base_char_masks_dict=dict([(char, source.get_activation(i, Network_UI.network['prediction_source', 0])) for i, char in enumerate(source.alphabet)])

            elif Network_UI.network['drum_act', 0] is not None: 
                source = Network_UI.network['drum_act'][0]
                self.base_char_masks_dict=dict([(source.instruments[char], source.get_activation(i, Network_UI.network['prediction_source', 0])) for i, char in enumerate(source.alphabet)])
            
            elif Network_UI.network['music_act', 0] is not None: 
                source = Network_UI.network['music_act'][0]
                self.base_char_masks_dict=dict([(source.midi_index_to_notestring(char), source.get_activation(i, Network_UI.network['prediction_source', 0])) for i, char in enumerate(source.alphabet)])
            

            Network_UI.Next_H_Block()

            self.sl0 = QSlider(1)
            self.sl0.setMinimum(0)
            self.sl0.setMaximum(100)
            self.sl0.setSliderPosition(50)
            self.sl0.label = QLabel('Size:')
            Network_UI.Add_element(self.sl0.label)
            Network_UI.Add_element(self.sl0, stretch=10)

            self.sl1 = QSlider(1)
            self.sl1.setMinimum(0)
            self.sl1.setMaximum(1000)
            self.sl1.setSliderPosition(100)
            self.sl1.label = QLabel('Speed:')
            Network_UI.Add_element(self.sl1.label)
            Network_UI.Add_element(self.sl1, stretch=10)

            self.sl2 = QSlider(1)
            self.sl2.setMinimum(0)
            self.sl2.setMaximum(3000)
            self.sl2.setSliderPosition(100)
            self.sl2.label = QLabel('Anti gravity:')
            Network_UI.Add_element(self.sl2.label)
            Network_UI.Add_element(self.sl2, stretch=10)

            self.sl3 = QSlider(1)
            self.sl3.setMinimum(0)
            self.sl3.setMaximum(100)
            self.sl3.setSliderPosition(0)
            self.sl3.label = QLabel('Noise:')
            Network_UI.Add_element(self.sl3.label)
            Network_UI.Add_element(self.sl3, stretch=10)

            self.sl4 = QSlider(1)
            self.sl4.setMinimum(0)
            self.sl4.setMaximum(500)
            self.sl4.setSliderPosition(100)
            self.sl4.label = QLabel('W^:')
            Network_UI.Add_element(self.sl4.label)
            Network_UI.Add_element(self.sl4, stretch=10)

            self.sl5 = QSlider(1)
            self.sl5.setMinimum(0)
            self.sl5.setMaximum(100)
            self.sl5.setSliderPosition(0)
            self.sl5.label = QLabel('Ring attraction:')
            Network_UI.Add_element(self.sl5.label)
            Network_UI.Add_element(self.sl5, stretch=10)

            self.weight_plot_cb = QCheckBox()
            self.weight_plot_cb.setText('Show weights')
            self.weight_plot_cb.setChecked(True)
            Network_UI.Add_element(self.weight_plot_cb, stretch=10)



    def update(self, Network_UI):
        if Network_UI.network['grammar_act', 0] is not None and self.sun_gravity_plot_tab.isVisible():
            #group = Network_UI.network['prediction_source', 0]

            groups = [Network_UI.network[tag, 0] for tag in Network_UI.neuron_visible_groups]

            self.draw_item.update_pic(groups, Network_UI.network['grammar_act', 0].alphabet, self.sl0.sliderPosition()/100, self.sl1.sliderPosition()/100, self.sl2.sliderPosition()/100, self.sl3.sliderPosition()/100, self.sl4.sliderPosition()/100, self.sl5.sliderPosition()/100, Network_UI, Network_UI.network['grammar_act', 0].get_char_input_statistics_list(), self.weight_plot_cb.isChecked())
            self.plot.update()

        elif Network_UI.network['drum_act', 0] is not None and self.sun_gravity_plot_tab.isVisible(): 
            groups = [Network_UI.network[tag, 0] for tag in Network_UI.neuron_visible_groups]
            source = Network_UI.network['drum_act', 0]
            alphabet = []
            for index in source.alphabet:
                alphabet.append(source.instruments[index])
            
            if source.include_inverse_alphabet: # include the symbols that stand for NOT the instrument
                for index in source.alphabet:
                    alphabet.append('! '+source.instruments[index])
            
            self.draw_item.update_pic(groups, alphabet, self.sl0.sliderPosition()/100, self.sl1.sliderPosition()/100, self.sl2.sliderPosition()/100, self.sl3.sliderPosition()/100, self.sl4.sliderPosition()/100, self.sl5.sliderPosition()/100, Network_UI, source.get_instrument_input_statistics_list(), self.weight_plot_cb.isChecked())
            self.plot.update()

        elif Network_UI.network['music_act', 0] is not None and self.sun_gravity_plot_tab.isVisible(): 
            groups = [Network_UI.network[tag, 0] for tag in Network_UI.neuron_visible_groups]
            source = Network_UI.network['music_act', 0]
            alphabet = []
            for index in source.alphabet:
                alphabet.append(source.midi_index_to_notestring(index))
            self.draw_item.update_pic(groups, alphabet, self.sl0.sliderPosition()/100, self.sl1.sliderPosition()/100, self.sl2.sliderPosition()/100, self.sl3.sliderPosition()/100, self.sl4.sliderPosition()/100, self.sl5.sliderPosition()/100, Network_UI, source.get_note_input_statistics_list(), self.weight_plot_cb.isChecked())
            self.plot.update()



from PymoNNto.NetworkBehaviour.Logic.TREN.Helper.Functions_Helper import *
from PymoNNto.NetworkBehaviour.Input.Activator import *
from PymoNNto.NetworkBehaviour.Logic.Basics.Normalization import *

class TREN_external_input(NeuronActivator):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.strength = self.get_init_attr('strength', 1.0, neurons)
        neurons.input = np.zeros(neurons.size)
        self.write_to = 'input'


    def new_iteration(self, neurons):
        neurons.input *= 0
        super().new_iteration(neurons)
        neurons.input *= self.strength


class InterGammaGlutamate(Behaviour):
    modificaton_reset_vars = ['']

    def set_variables(self, neurons):
        self.add_tag('Inter GLU')
        neurons.glu_inter_gamma_activity = neurons.get_neuron_vec()
        neurons.require_synapses('GLU')


    def new_iteration(self, neurons):
        neurons.glu_inter_gamma_activity *= 0

        for s in neurons.afferent_synapses['GLU']:
            s.slow_add=np.dot(s.W, s.src.output_activity_history[0])
            s.dst.glu_inter_gamma_activity+=s.slow_add
            #s.get_dest_vec('glu_inter_gamma_activity')[:] += np.dot(s.W, s.get_src_vec_obj(s.src.output_activity_history[0])[:])

    def get_shared_variable(self, name):
        if name == 'buffer_size':
            return 1



class IntraGammaGlutamate(Behaviour):
    def set_variables(self, neurons):
        self.add_tag('Intra GLU')
        neurons.glu_intra_gamma_activity = neurons.get_neuron_vec()

        neurons.require_synapses('GLU')


    def new_iteration(self, neurons):
        neurons.glu_intra_gamma_activity *= 0
        for s in neurons.afferent_synapses['GLU']:
            if hasattr(s.src, 'glu_inter_gamma_activity'):
                s.fast_add=np.dot(s.W,  relu3(s.src.glu_inter_gamma_activity, neurons.TH,  0))
                s.dst.glu_intra_gamma_activity += s.fast_add
                #s.get_dest_vec('glu_intra_gamma_activity')[:] += np.dot(s.W,  relu3(s.get_src_vec('glu_inter_gamma_activity')[:], neurons.TH,  0))


class IntraGammaGABA(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('Intra GABA')
        self.GABA_density = self.get_init_attr('GABA_density', 1.0, neurons)
        self.GABA_random_factor = self.get_init_attr('GABA_random_factor', 0.0, neurons)
        self.GABA_Norm = self.get_init_attr('GABA_Norm', 4.3, neurons)

        all_neurons_same = self.get_init_attr('all_neurons_same', False, neurons)

        neurons.gaba_intra_gamma_activity = neurons.get_neuron_vec()

        neurons.require_synapses('GABA')

        self.initialize_synapse_attr('W', self.GABA_density, 1.0, self.GABA_random_factor, neurons, 'GABA', all_neurons_same)

        normalize_synapse_attr('W', 'W', self.GABA_Norm, neurons, 'GABA')




    def new_iteration(self, neurons):
        neurons.gaba_intra_gamma_activity *= 0
        for s in neurons.afferent_synapses['GABA']:
            s.dst.gaba_intra_gamma_activity -= np.dot(s.W, relu3(s.src.glu_inter_gamma_activity, neurons.TH, 0))
            #s.get_dest_vec('gaba_intra_gamma_activity')[:] -= np.dot(s.GABA_Synapses, relu3(s.get_src_vec('glu_inter_gamma_activity')[:],neurons.TH,0))


    def get_shared_variable(self, name):
        if name == 'buffer_size':
            return 1


class ActivityBuffering(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('Collect and Buffer')
        #self.store_input = self.get_init_attr('store_input', True)
        self.min_buffersize = self.get_init_attr('min_buffersize', 2, neurons)
        self.activity_multiplyer = self.get_init_attr('activity_multiplyer', 0.0, neurons)
        neurons.TH = self.get_init_attr('firetreshold', 0.1, neurons)

        buffersize = np.max(neurons.get_shared_variables('buffer_size'))#problem:2 verschiedene gruppen

        neurons.activity = neurons.get_neuron_vec()
        neurons.output_activity_history = neurons.get_neuron_vec_buffer(buffersize)


    def new_iteration(self, neurons):

        neurons.activity *= self.activity_multiplyer
        neurons.activity += neurons.glu_inter_gamma_activity.copy()

        #experimetal...
        neurons.pre_inhibition_act = neurons.activity.copy()

        if hasattr(neurons, 'glu_intra_gamma_activity'):
            neurons.activity += neurons.glu_intra_gamma_activity
        if hasattr(neurons, 'gaba_intra_gamma_activity'):
            neurons.activity += neurons.gaba_intra_gamma_activity
        if hasattr(neurons, 'random_leak_activity'):
            neurons.activity += neurons.random_leak_activity


        neurons.activity = np.clip(neurons.activity, 0, 1)

        #if self.store_input:
        #    neurons.input_activity_history = roll(neurons.input_activity_history)#np.roll(neurons.input_activity_history, 1, axis=0)
        #    neurons.input_activity_history[0] = neurons.activity

        neurons.output_activity_history = roll(neurons.output_activity_history)#np.roll(neurons.output_activity_history, 1, axis=0)
        neurons.output = np.copy(relu3(neurons.activity+neurons.input, neurons.TH, 0.0))
        neurons.output_activity_history[0] = neurons.output#0.1#
        #neurons.output_activity_history[0] = np.copy(neurons.activity)

    def get_shared_variable(self, name):
        if name == 'buffer_size':
            return self.min_buffersize


class RandomLeakInput(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('Random Leak')
        neurons.random_leak_activity = neurons.get_neuron_vec()
        self.random_strength = self.get_init_attr('random_strength', 1, neurons)

    def new_iteration(self, neurons):
        neurons.random_leak_activity = neurons.get_random_neuron_vec()*neurons.weight_norm_factor*self.random_strength

class additional(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('additional ...')
        neurons.output_new = neurons.get_neuron_vec()
        neurons.excitation = neurons.get_neuron_vec()
        neurons.inhibition = neurons.get_neuron_vec()
        neurons.input_act = neurons.get_neuron_vec()
        neurons.refractory_counter = neurons.get_neuron_vec()
        neurons.nox = neurons.get_neuron_vec()

        for s in neurons.afferent_synapses['GLU']+neurons.afferent_synapses['GABA']:
            s.slow_add = neurons.get_neuron_vec()
            s.fast_add = neurons.get_neuron_vec()


    def new_iteration(self, neurons):
        return




from PymoNNto.NetworkCore.Base import *

class PatternGroup(NetworkObjectBase):

    def __init__(self, **kwargs):
        super().__init__(kwargs.get('tag', None))
        self.kwargs=kwargs

        self.active = True
        self.current_pattern_index = -1

        if kwargs is not None:
            self.group_possibility = kwargs.get('group_possibility', 1.0)
            self.overlapping_with_other_groups = kwargs.get('overlapping_with_other_groups', False)
            self.has_overlapping_priority = kwargs.get('has_overlapping_priority', True)

        self.initialize()

    def initialize(self):
        return

    def initialize_neuron_group(self, neurons):
        return

    def get_pattern(self, neurons):
        return None




from PymoNNto.Exploration.Network_UI.TabBase import *

#from Testing.Common.Classifier_Helper import *

class sidebar_music_module(TabBase):

    def add_recorder_variables(self, neuron_group, Network_UI):
        if hasattr(neuron_group, 'pattern_index'):
            Network_UI.add_recording_variable(neuron_group, 'n.pattern_index', timesteps=100)

    def initialize(self, Network_UI):
        if len(Network_UI.network['music_act'])>0:

            self.readout = None

            def learning_on_off(event):
                Network_UI.network.set_mechanisms(['STDP'], self.stdp_cb.isChecked())

            network = Network_UI.network
            source = network['music_act', 0]
            min_midi = min(filter(lambda x: x > 0, source.alphabet))
            max_midi = max(source.alphabet)

            stringaxis = pg.AxisItem(orientation='left')

            ticks = np.arange(24, 121, 12)
            ticks = ticks[np.where(np.logical_and(ticks>=min_midi, ticks<=max_midi))]
            ticks_strings= []
            for i in range(len(ticks)):
                ticks_strings.append(source.midi_index_to_notestring(ticks[i]))
            c_ticks = dict(zip(ticks,ticks_strings))
            stringaxis.setTicks([c_ticks.items()])

            p = Network_UI.Add_plot(sidebar=True, title='Input', axisItems={'left': stringaxis})

            self.scatter = pg.ScatterPlotItem()
            p.showGrid(y=True)
            p.addItem(self.scatter)

            p.setYRange(min_midi,max_midi, padding=0)
            self.n_notes_shown=20
            p.setXRange(-1,self.n_notes_shown+1)

            #image = pg.ImageItem(np.rot90(np.asarray(Image.open('midis_piano_cropped.png'))))
            #image.scale(0.05, 0.1)
            #p.addItem(image)

    
    def update(self, Network_UI):
        if len(Network_UI.network['music_act'])>0:

            network = Network_UI.network
            source = network['music_act', 0]
            curr_output = np.array(network['n.pattern_index', 0][-self.n_notes_shown:])

            if source.type=='poly':
                y=[]
                x=[]
                ts=0
                for i in range(curr_output.shape[0]):
                    indices = np.nonzero(curr_output[i])[0]
                    if len(indices)==0:
                        x.append(ts)
                        y.append(np.nan)
                    else:
                        for j in indices:
                            x.append(ts)
                            y.append(source.alphabet[j])
                    ts+=1
                self.scatter.setData(x,y)

            elif source.type=='mono':
                midi_indices = np.empty(curr_output.shape)
            
                for i in range(len(curr_output)):
                    midi_indices[i] = source.alphabet[curr_output[i]]
                    if midi_indices[i] == -1:
                        midi_indices[i]=np.nan
                    
                x = np.arange(len(midi_indices))
                y = midi_indices

                #if len(midi_indices)>=self.n_notes_shown:
            if len(x) > 0:#self.n_notes_shown:
                self.scatter.setData(x,y)
                self.scatter.setSize(2)

            #current_corpus_index = source.current_index()
        



from PymoNNto.NetworkBehaviour.Input.Activator import *
from PymoNNto.NetworkBehaviour.Logic.Basics.Normalization import *

##########################################################################
#Helper functions
##########################################################################
def modular_activation_function(neurons):
    result = neurons.activity.copy()

    if hasattr(neurons, 'rnd_act_factor') and neurons.rnd_act_factor is not None:
        result += neurons.rnd_act_factor * np.random.rand(neurons.size)

    if hasattr(neurons, 'nox') and neurons.nox is not None:
        result -= neurons.nox

    if hasattr(neurons, 'refractory_counter_analog') and neurons.refractory_counter_analog is not None:
        result -= neurons.refractory_counter_analog

    if hasattr(neurons, 'TH') and neurons.TH is not None:
        result -= neurons.TH

    if hasattr(neurons, 'digital_output') and neurons.digital_output:
        result = (result >= 0.0).astype(def_dtype)

    if hasattr(neurons, 'refractory_counter_digital') and neurons.refractory_counter_digital is not None:
        result *= (neurons.refractory_counter_digital < neurons.refractory_counter_threshold)

    return result

def last_cycle_step(neurons):
    return neurons.iteration % neurons.timescale == neurons.timescale-1

def first_cycle_step(neurons):
    return neurons.iteration % neurons.timescale == 0



##########################################################################
#Init synapses
##########################################################################
class SORN_init_afferent_synapses(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('init_afferent_synapses')

        #add synapsegroup.group_weighting attribute to change its weight compared to synapsegroups of same kind

        self.transmitter = self.get_init_attr('transmitter', None, neurons)

        density = self.get_init_attr('density', None, neurons)
        if type(density)==str and density[-1] == '%':
            density=float(density[:-1])/100

        density_mode = self.get_init_attr('density_mode', 'auto', neurons)

        self.distribution = self.get_init_attr('distribution', 'random_sample(dim)', neurons, do_not_diversify=True)

        if self.distribution is not None and ';plot' in self.distribution:
            import matplotlib.pyplot as plt
            self.distribution = self.distribution.replace(';plot', '')
            plt.hist(eval(self.distribution.replace(')', ',size=1000)')), bins=100)
            plt.show()

        partition_compensation = self.get_init_attr('partition_compensation', False, neurons)

        # np.random. ...
        #lognormal([mean, sigma, size])
        #normal([loc, scale, size])
        #random_sample(dim)
        #https://docs.scipy.org/doc/numpy-1.13.0/reference/routines.random.html


        if density == 'full':
            density_mode = 'full'
            density = 1

        if density_mode == 'auto':
            if np.min(density) < 1: #min because of array possibility
                density_mode = 'chance'
            else:
                density_mode = 'syn_count'


        if density_mode == 'full':
            density = 1
        elif density_mode == 'chance':
            density = density
        elif density_mode == 'syn_count':
            base_density = density#...
        elif density_mode == 'fixed_syn_count':
            density = -1
            #todo:implement (difficult with s.enabled...)
            #s.get_random_synapse_mat_fixed(syn_lamb)
        else:
            print('invalid density mode')

        self.group_weighting = self.get_init_attr('group_weighting', False, neurons)#each group can be weighted individually with a unique parameter (s.group_weighting)

        for s in neurons.afferent_synapses[self.transmitter]:
            if self.group_weighting == False:
                weighting = 1
            else:
                weighting = s.get_synapse_group_size_factor(s, self.transmitter)#todo test

            if partition_compensation:
                weighting *= s.src.group_without_subGroup().size/s.src.size

            if density_mode == 'syn_count':
               density = base_density/s.src.size

            if self.distribution is not None:
                s.W = s.get_random_synapse_mat(density=float(density*weighting), rnd_code=self.distribution)#density*weighting
            else:
                s.W = (s.get_synapse_mat()+1)*s.enabled #all same

            #remove_auto_synapses
            if s.src==s.dst:
                for d in range(s.dst.size):
                    s.W[d, d] = 0.0

            s.enabled *= (s.W > 0)

            s.slow_add=0
            s.fast_add=0

        self.normalize = self.get_init_attr('normalize', True, neurons)# can be number or bool
        if self.normalize is not None and self.normalize is not False:
            if self.normalize==True:
                self.normalize=1.0
            normalize_synapse_attr('W', 'W', self.normalize, neurons, self.transmitter)

        #for s in neurons.afferent_synapses[self.transmitter]:
        #    plt.hist(s.W.flatten()[s.W.flatten()>0],bins=100)
        #    plt.show()

        self.clip_min = self.get_init_attr('clip_min', 0.0, neurons)
        self.clip_max = self.get_init_attr('clip_max', None, neurons)


    def new_iteration(self, neurons):
        for s in neurons.afferent_synapses[self.transmitter]:
            s.W = s.W*s.enabled

        if self.clip_min is not None or self.clip_max is not None:
            s.W = np.clip(s.W, self.clip_min, self.clip_max)




class SORN_temporal_synapses(Behaviour):
    def set_variables(self, neurons):
        self.add_tag('temporal_synapses')

        self.syn_type = self.get_init_attr('syn_type', 1)
        self.shift_factor = self.get_init_attr('shift_factor', 0.1)
        self.shift_loss_factor = self.get_init_attr('shift_loss_factor', 0.5)

        for s in neurons.afferent_synapses[self.syn_type]:
            s.W_stable = s.W.copy()
            s.W_temp = s.get_synapse_mat()

        self.behaviour_norm_factor = self.get_init_attr('behaviour_norm_factor', 1.0, neurons)
        neurons.weight_norm_factor = neurons.get_neuron_vec()+self.get_init_attr('neuron_norm_factor', 1.0, neurons)

    def new_iteration(self, neurons):

        if last_cycle_step(neurons):
            for s in neurons.afferent_synapses[self.syn_type]:
                #s.W_temp += (s.get_synapse_mat()+0.15)*(s.get_random_synapse_mat()>0.99)
                s.W = s.W_stable + s.W_temp
                s.W[s.W < 0.0] = 0.0

            normalize_synapse_attr('W', 'W_temp', neurons.weight_norm_factor * self.behaviour_norm_factor, neurons, self.syn_type)
            normalize_synapse_attr('W', 'W_stable', neurons.weight_norm_factor * self.behaviour_norm_factor, neurons, self.syn_type)

            for s in neurons.afferent_synapses[self.syn_type]:
                shift = s.W_temp * self.shift_factor
                s.W_temp -= shift
                s.W_stable += shift*self.shift_loss_factor



class SORN_dopamine(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('temporal_synapses')

        self.temporal_synapse_tag = self.get_init_attr('temporal_synapse_tag', 'GLU')
        self.positive_dopamine_tag = self.get_init_attr('dopamine_tag', 'DOP+')
        self.negative_dopamine_tag = self.get_init_attr('dopamine_tag', 'DOP-')

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            if self.positive_dopamine_tag in neurons.afferent_synapses:
                positive_dopamine_level = np.mean([np.mean(s.src.output) for s in neurons.afferent_synapses[self.positive_dopamine_tag]])
                for s in neurons.afferent_synapses[self.temporal_synapse_tag]:
                    positive_w_shift = s.W_temp * positive_dopamine_level
                    s.W_temp -= positive_w_shift  # w_temp_gets_smaller (neccessary???)
                    s.W_stable += positive_w_shift #W_stable gets larger

            if self.negative_dopamine_tag in neurons.afferent_synapses:
                negative_dopamine_level = np.mean([np.mean(s.src.output) for s in neurons.afferent_synapses[self.negative_dopamine_tag]])
                for s in neurons.afferent_synapses[self.temporal_synapse_tag]:
                    negative_w_shift = s.W_temp * negative_dopamine_level
                    s.W_temp -= negative_w_shift  # w_temp_gets_smaller (neccessary???)
                    s.W_stable -= negative_w_shift #W_stable gets smaller



##########################################################################
#Init neurons
##########################################################################
class SORN_init_neuron_vars(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('init_neuron_vars')

        neurons.activity = neurons.get_neuron_vec()
        neurons.excitation = neurons.get_neuron_vec()
        neurons.inhibition = neurons.get_neuron_vec()
        neurons.input_act = neurons.get_neuron_vec()

        neurons.output = neurons.get_neuron_vec()

        #neurons.slow_act=neurons.get_neuron_vec()

        neurons.timescale = self.get_init_attr('timescale', 1)

    def new_iteration(self, neurons):
        if first_cycle_step(neurons):
            neurons.activity.fill(0)# *= 0
            neurons.excitation.fill(0)# *= 0
            neurons.inhibition.fill(0)# *= 0
            neurons.input_act.fill(0)# *= 0

            neurons.output_old=neurons.output.copy()

        #neurons.buffer_timescale_sum_dict={}

            #neurons.slow_act *= 0


##########################################################################
#NOX
##########################################################################
class SORN_NOX(Instant_Homeostasis):

    def partition_sum(self, neurons):
        neurons._temp_act_sum = neurons.get_neuron_vec()
        for s in neurons.afferent_synapses['GLU']:
            s.dst._temp_act_sum += np.mean(s.src.output)
        return neurons._temp_act_sum

    def set_variables(self, neurons):
        self.add_tag('NOX')
        super().set_variables(neurons)
        neurons.nox = neurons.get_neuron_vec()
        self.measurement_param = self.get_init_attr('mp', 'self.partition_sum(n)', None)#'np.mean(n.output_new)'
        self.adjustment_param = 'nox'
        self.measure_group_sum = True

        self.set_threshold(self.get_init_attr('h_dh', 0, neurons))#'same(IPTI, th)'
        self.adj_strength = -self.get_init_attr('eta_nox', 0.002, neurons)

        self.distance_sensitive=True

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            neurons.nox.fill(0)# *= 0
            super().new_iteration(neurons)


##########################################################################
#Basic activity transission class
##########################################################################
class SORN_signal_propagation_base(Behaviour):

    def set_variables(self, neurons):
        self.transmitter = self.get_init_attr('transmitter', None, neurons)
        self.strength = self.get_init_attr('strength', 1, neurons)  # 1 or -1

        if self.transmitter==None:
            print('warning no transmitter defined')
        #elif self.transmitter is not 'GLU' and self.transmitter is not 'GABA':
        #    print('warning unknown transmitter')

        if self.transmitter=='GLU' and self.strength<0:
            print('warning glutamate strength is inhibitory')

        if self.transmitter=='GABA' and self.strength>0:
            print('warning GABA strength is excitatory')

    def new_iteration(self, neurons):
        print('warning: signal_propagation_base has to be overwritten')


##########################################################################
#External Input
##########################################################################
class SORN_external_input(NeuronActivator):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.strength = self.get_init_attr('strength', 1.0, neurons)
        neurons.input = np.zeros(neurons.size)
        self.write_to = self.get_init_attr('write_to', 'input', neurons)
        neurons.add_tag('text_input_group')
        #pre_syn = neurons.connected_NG_param_list('afferent_buffer_requirement', same_NG=True, search_behaviours=True)
        neurons.input_buffer = neurons.get_neuron_vec_buffer(neurons.timescale)#pre_syn[0][0]


    def new_iteration(self, neurons):
        super().new_iteration(neurons)

        neurons.input_buffer = neurons.buffer_roll(neurons.input_buffer, neurons.input)

        if last_cycle_step(neurons) and self.strength != 0:
            add = np.sum(neurons.input_buffer[neurons.timescale-1:neurons.timescale*2-1], axis=0)/neurons.timescale * self.strength#neurons.input * self.strength / neurons.timescale
            neurons.activity += add
            neurons.input.fill(0)# *= 0
            neurons.input_act += add


##########################################################################
#Slow(normal) transmission
##########################################################################
class SORN_slow_syn(SORN_signal_propagation_base):

    def afferent_buffer_requirement(self, neurons):
        return buffer_reqirement(length=1, variable='output', timescale=neurons.timescale)
        #return 1, neurons.timescale#, 0

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.add_tag('slow_' + self.transmitter)
        self.sparse_optimization = self.get_init_attr('so', False, neurons)

        self.pre_syn_groups = []
        for s in neurons.afferent_synapses[self.transmitter]:
            if s.src.BaseNeuronGroup not in self.pre_syn_groups:
                self.pre_syn_groups.append(s.src.BaseNeuronGroup)


    def new_iteration(self, neurons):
        if last_cycle_step(neurons) and self.strength != 0:

            for s in neurons.afferent_synapses[self.transmitter]:

                #mask = True
                #if hasattr(s.src, 'temporal_shift_mask'):
                #    mask = s.src.temporal_shift_mask == (s.src.iteration % s.src.timescale)

                #*mask
                s.slow_add = s.W.dot(get_buffer(s.src, 'output', neurons.timescale, 0)[0]) * self.strength

                s.dst.activity += s.slow_add
                if self.strength > 0:
                    s.dst.excitation += s.slow_add
                else:
                    s.dst.inhibition += s.slow_add


##########################################################################
#Fast(same clock cycle) transmission
##########################################################################
class SORN_fast_syn(SORN_signal_propagation_base):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.add_tag('fast_' + self.transmitter)
        self.sparse_optimization = self.get_init_attr('so', False, neurons)

    def new_iteration(self, neurons):
        if last_cycle_step(neurons) and self.strength!=0:
            for s in neurons.afferent_synapses[self.transmitter]:
                #if self.sparse_optimization:
                #    s.fast_add = (np.sum(s.W[:, s.src.activation_function(s.src).astype(np.bool)], axis=1) * self.strength) / neurons.timescale
                #    #add = (np.sum(s.W.transpose()[s.src.output.astype(np.bool)], axis=0) * self.strength) / neurons.timescale
                #else:

                mask = True
                if hasattr(s.src, 'temporal_shift_mask'):
                    mask = s.src.temporal_shift_mask == (s.src.iteration % s.src.timescale)

                s.fast_add = s.W.dot(s.src.activation_function(s.src) * mask) * self.strength# / neurons.timescale

                s.dst.activity += s.fast_add
                if self.strength > 0:
                    s.dst.excitation += s.fast_add
                else:
                    s.dst.inhibition += s.fast_add


##########################################################################
#Generate output with activation fucntion
##########################################################################
class SORN_generate_output(Behaviour):#has to be executed AFTER intra, inter...behaviours

    def set_variables(self, neurons):
        self.add_tag('generate output')

        neurons.output = neurons.get_neuron_vec()

        neurons.digital_output = self.get_init_attr('digital_output', True, neurons)

        neurons.activation_function = self.get_init_attr('activation_function', modular_activation_function, neurons)
        if not callable(neurons.activation_function):
            neurons.activation_function = eval(neurons.activation_function)

        neurons.TH = neurons.get_neuron_vec()+self.get_init_attr('init_TH', 0.0, neurons)#neurons.get_random_neuron_vec()*0.5

        neurons.rnd_act_factor = self.get_init_attr('rnd_act_factor', None, neurons)  # sigma

        #neurons.get_buffer = get_buffer

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            neurons.output = neurons.activation_function(neurons)


##########################################################################
#Buffer allocation and output production
##########################################################################
class buffer_reqirement:
    def __init__(self, length=-1, variable='', timescale=1, offset=0):
        self.length = length
        self.variable = variable
        self.timescale = timescale
        self.offset = offset

def get_buffer(neurons, variable, timescale=1, offset=0):
    return neurons.mask_var(neurons.buffers[variable][timescale][offset])

class SORN_buffer_variables(Behaviour):#has to be executed AFTER intra, inter...behaviours

    def set_variables(self, neurons):
        n=neurons#for compile...
        self.add_tag('buffer')
        neurons.get_buffer = get_buffer#just for simpler access

        post_syn_req = neurons.connected_NG_param_list('afferent_buffer_requirement', syn_tag='All', efferent_NGs=True, search_behaviours=True)
        own_req = neurons.connected_NG_param_list('own_buffer_requirement', same_NG=True, search_behaviours=True)
        pre_syn_req = neurons.connected_NG_param_list('efferent_buffer_requirement', syn_tag='All', afferent_NGs=True, search_behaviours=True)
        all_requirments = post_syn_req+own_req+pre_syn_req

        variable_key = np.unique([vr.variable+'_'+str(vr.timescale)+'_'+str(vr.offset) for vr in all_requirments]) #remove timescale

        neurons.precompiled_vars = {}
        neurons.buffers = {}

        def add_buffer(variable, timescale, offset, length):#create dict(offset) in a dict(timescale) in a dict(variable)...
            #print(n.tags, variable, timescale, offset, length)

            if '.' in variable or '(' in variable:
                neurons.precompiled_vars[variable] = compile(variable, '<string>', 'eval')

            if variable not in neurons.buffers:
                neurons.buffers[variable] = {}
            if timescale not in neurons.buffers[variable]:
                neurons.buffers[variable][timescale] = {}

            current = None
            if offset in neurons.buffers[variable][timescale]:#buffer already exists:
                current=len(neurons.buffers[variable][timescale][offset])

            if current==None or length>current:
                neurons.buffers[variable][timescale][offset]=neurons.get_neuron_vec_buffer(length)

        for var_key in variable_key:
            split_key=var_key.split('_')
            variable = split_key[0]
            timescale = int(split_key[1])
            offset = int(split_key[2])
            max_req_length = np.max([vr.length for vr in all_requirments if vr.variable == variable and vr.timescale == timescale and vr.offset == offset])
            add_buffer(variable, 1, 0, timescale)#??????
            add_buffer(variable, timescale, offset, max_req_length)


        ####################################################

        if self.get_init_attr('random_temporal_output_shift', True) and neurons.timescale > 1:
            neurons.temporal_shift_mask = np.random.randint(neurons.timescale, size=neurons.size)


    def new_iteration(self, neurons):
        n = neurons  # for compile...

        for variable in neurons.buffers:
            for timescale in neurons.buffers[variable]:
                if timescale == 1:
                    if variable in neurons.precompiled_vars:
                        new = eval(variable)
                    else:
                        new = getattr(neurons, variable)
                    neurons.buffer_roll(neurons.buffers[variable][1][0], new)

        for variable in neurons.buffers:
            for timescale in neurons.buffers[variable]:
                if timescale > 1:
                    for offset in neurons.buffers[variable][timescale]:
                        if neurons.iteration % timescale == timescale - 1 - offset:
                            if timescale <= neurons.timescale:
                                new = neurons.buffers[variable][1][0][0]#previously saved value
                                if hasattr(neurons, 'temporal_shift_mask'):
                                    new *= neurons.temporal_shift_mask == (neurons.iteration % neurons.timescale)
                            else:
                               new = np.mean(neurons.buffers[variable][1][0][0:timescale], axis=0)#sum of previously saved value

                            neurons.buffer_roll(neurons.buffers[variable][timescale][offset], new)

                            #if variable in neurons.precompiled_vars:
                            #    new = eval(variable)
                            #else:
                            #    new = getattr(neurons, variable)




        '''
        for variable in neurons.buffers:
            for timescale in neurons.buffers[variable]:
                if timescale == 1:
                    if variable in neurons.precompiled_vars:
                        new = eval(variable)
                    else:
                        new = getattr(neurons, variable)
                    neurons.buffer_roll(neurons.buffers[variable][1][0], new)

        for variable in neurons.buffers:
            for timescale in neurons.buffers[variable]:
                if timescale > 1:
                    for offset in neurons.buffers[variable][timescale]:
                        if neurons.iteration % timescale == timescale - 1 - offset:
                            data = np.mean(neurons.buffers[variable][1][offset][0:timescale], axis=0)
                            neurons.buffer_roll(neurons.buffers[variable][timescale][offset], data)
        '''

class SORN_STDP_simple(Behaviour):
    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.add_tag('SORN_IP_WTA_apply')
        neurons.eta_stdp = self.get_init_attr('eta_stdp', 0.00015, neurons)
        self.syn_type = self.get_init_attr('syn_type', 'GLU', neurons)

    def new_iteration(self, neurons):
        for s in neurons.afferent_synapses[self.syn_type]:
            pre_post = s.dst.output[:, None] * s.src.output_old[None, :]
            simu = s.dst.output[:, None] * s.src.output[None, :]
            post_pre = s.dst.output_old[:, None] * s.src.output[None, :]

            dw = neurons.eta_stdp * (pre_post - post_pre +simu)

            # dw = neurons.eta_stdp * (pre_post - post_pre + simu)

            # dw = neurons.eta_stdp * (to_new[:, None] * from_old[None, :] - to_old[:, None] * from_new[None, :])

            s.W=np.clip(s.W + dw, 0.0, 1.0)

class SORN_STDP(Behaviour):

    def get_STDP_Function(self):
        return self.get_init_attr('STDP_F', {-1: 1, 1: -1})

    def afferent_buffer_requirement(self, neurons):
        self.STDP_F = self.get_STDP_Function()
        self.data = np.array([[t, s] for t, s in self.STDP_F.items()])
        length = int(np.maximum(np.max(self.data[:, 0]), 1)+1)
        #print('aff', length, self.get_init_attr('STDP_F', {-1: 1, 1: -1}))
        return buffer_reqirement(length=length, variable='output', timescale=neurons.timescale)
        #return int(np.maximum(np.min(self.data[:, 0])*-1, 1)+1), neurons.timescale#, 0

    def own_buffer_requirement(self, neurons):
        self.STDP_F = self.get_STDP_Function()
        self.data = np.array([[t, s] for t, s in self.STDP_F.items()])
        length = int(np.maximum(np.min(self.data[:, 0])*-1, 1)+1)
        #print('own',length, self.get_init_attr('STDP_F', {-1: 1, 1: -1}))
        return buffer_reqirement(length=length, variable='output', timescale=neurons.timescale)
        #return int(np.maximum(np.max(self.data[:, 0]), 1)+1), neurons.timescale#, 0

    def set_variables(self, neurons):
        self.add_tag('STDP')

        self.weight_attr = self.get_init_attr('weight_attr', 'W', neurons)

        self.STDP_F = self.get_STDP_Function()# left(negative t):pre->post right(positive t):post->pre

        self.pre_post_mask = np.array([t in self.STDP_F for t in range(self.afferent_buffer_requirement(neurons).length)])
        self.pre_post_mul = np.array([self.STDP_F[t] for t in range(self.afferent_buffer_requirement(neurons).length) if t in self.STDP_F])

        self.post_pre_mask = np.array([-t in self.STDP_F for t in range(self.own_buffer_requirement(neurons).length)])
        self.post_pre_mul = np.array([self.STDP_F[-t] for t in range(self.own_buffer_requirement(neurons).length) if -t in self.STDP_F])

        self.transmitter = self.get_init_attr('transmitter', 'GLU', neurons)

        neurons.eta_stdp = self.get_init_attr('eta_stdp', 0.005, neurons)
        neurons.last_output = neurons.get_neuron_vec()
        for s in neurons.afferent_synapses[self.transmitter]:
            s.src_act_sum_old = np.zeros(s.src.size)

        if self.get_init_attr('plot', False):
            import matplotlib.pyplot as plt
            self.data = np.array([[x, y] for x, y in self.STDP_F.items()])
            plt.bar(self.data[:, 0], self.data[:, 1], 1.0)
            plt.axhline(0, color='black')
            plt.axvline(0, color='black')
            plt.show()

    def new_iteration(self, neurons):
        #if first_cycle_step(neurons):#first????????????????????????????????????????????????????????????????????????
        if last_cycle_step(neurons):

            for s in neurons.afferent_synapses[self.transmitter]:

                post_act = get_buffer(s.dst, 'output', neurons.timescale)#s.dst.get_masked_dict('output_buffer_dict', neurons.timescale)
                pre_act = get_buffer(s.src, 'output', neurons.timescale)#s.src.get_masked_dict('output_buffer_dict', neurons.timescale)

                if not hasattr(s, 'pre_post_mask'):
                    s.pre_post_mask = self.pre_post_mask.copy()
                    if len(post_act) > len(s.pre_post_mask):
                        s.pre_post_mask = np.concatenate([s.pre_post_mask, np.array([False for _ in range(len(post_act) - len(s.pre_post_mask))])])

                if not hasattr(s, 'post_pre_mask'):
                    s.post_pre_mask = self.post_pre_mask.copy()
                    if len(pre_act) > len(s.post_pre_mask):
                        s.post_pre_mask = np.concatenate([s.post_pre_mask, np.array([False for _ in range(len(pre_act) - len(s.post_pre_mask))])])

                summed_up_dact = np.sum(post_act[s.pre_post_mask]*self.pre_post_mul[:, None], axis=0)
                summed_up_sact = np.sum(pre_act[s.post_pre_mask]*self.post_pre_mul[:, None], axis=0)

                dw_pre_post = summed_up_dact[:, None] * pre_act[0][None, :]
                dw_post_pre = post_act[0][:, None] * summed_up_sact[None, :]

                s.dw = neurons.eta_stdp * (dw_pre_post+dw_post_pre)# * (neurons.timescale)

                setattr(s, self.weight_attr, getattr(s, self.weight_attr)+s.dw)

                #W = getattr(s, self.weight_attr)[:]
                #W += s.dw
                #W[W < 0.0] = 0.0


class SORN_Refractory_Digital(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('Refractory_D')
        neurons.refractory_counter_digital = neurons.get_neuron_vec()
        self.factor = self.get_init_attr('factor', 0.9, neurons)
        neurons.refractory_counter_threshold = self.get_init_attr('threshold', 0.1, neurons)

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            neurons.refractory_counter_digital *= self.factor
            neurons.refractory_counter_digital += neurons.output


class SORN_Refractory_Analog(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('Refractory_A')
        neurons.refractory_counter_analog = neurons.get_neuron_vec()
        self.factor = self.get_init_attr('factor', 0.9, neurons)

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            neurons.refractory_counter_analog *= self.factor
            neurons.refractory_counter_analog += neurons.output


class SORN_SN(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('SN')
        self.syn_type = self.get_init_attr('syn_type', 'GLU', neurons)

        neurons.require_synapses(self.syn_type, warning=False)#suppresses error when synapse group does not exist

        #self.clip_min = self.get_init_attr('clip_min', None, neurons)
        #self.clip_max = self.get_init_attr('clip_max', None, neurons)

        self.only_positive_synapses = self.get_init_attr('only_positive_synapses', True, neurons)
        #if self.only_positive_synapses:
        #    self.clip_min = self.get_init_attr('clip_min', 0.0, neurons)

        self.behaviour_norm_factor = self.get_init_attr('behaviour_norm_factor', 1.0, neurons)
        neurons.weight_norm_factor = neurons.get_neuron_vec()+self.get_init_attr('neuron_norm_factor', 1.0, neurons)

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            if self.only_positive_synapses:
                for s in neurons.afferent_synapses[self.syn_type]:
                    s.W[s.W < 0.0] = 0.0

            normalize_synapse_attr('W', 'W', neurons.weight_norm_factor*self.behaviour_norm_factor, neurons, self.syn_type)

            #if self.clip_max is not None:
            #    for s in neurons.afferent_synapses[self.syn_type]:
            #        s.W[s.W > self.clip_max] = self.clip_max

            #if self.clip_min is not None:
            #    for s in neurons.afferent_synapses[self.syn_type]:
            #        s.W[s.W < self.clip_min] = self.clip_min

                #s.W = np.clip(s.W, 0, self.clip_max)


class SORN_iSTDP(Behaviour):

    def afferent_buffer_requirement(self, neurons):
        return buffer_reqirement(length=1, variable='output', timescale=neurons.timescale)

    def own_buffer_requirement(self, neurons):
        return buffer_reqirement(length=1, variable='n.excitation+n.inhibition', timescale=neurons.timescale)

    def set_variables(self, neurons):
        self.add_tag('iSTDP')

        self.transmitter = self.get_init_attr('transmitter', 'GABA', neurons)

        neurons.eta_istdp = self.get_init_attr('eta_istdp', 0.001, neurons)
        neurons.h_ip = self.get_init_attr('h_ip', 0.1, neurons)

        #for s in neurons.afferent_synapses[self.transmitter]:
        #    s.iSTDP_src_lag_buffer = np.zeros(s.src.size)

    def new_iteration(self, neurons):
        for s in neurons.afferent_synapses[self.transmitter]:
            #s.iSTDP_src_lag_buffer += s.src.output/neurons.timescale
            if last_cycle_step(neurons):

                s.W += -s.dst.eta_istdp * ((1 - get_buffer(s.dst, 'n.excitation+n.inhibition', neurons.timescale, 0)[0] * (1 + 1.0 / s.dst.h_ip))[:, None] * get_buffer(s.src, 'output', neurons.timescale, 0)[0])#(s.dst.excitation+s.dst.inhibition)#s.iSTDP_src_lag_buffer[None, :]
                s.W = np.clip(s.W, 0.001, 1.0)

                #s.iSTDP_src_lag_buffer.fill(0)# *= 0

                #dW_same = -self.eta_istdp * ((1 - (s.dst.output[:, None] * (1 + 1.0 / self.h_ip))) * s.src.output[None, :])


class SORN_SC_TI(Time_Integration_Homeostasis):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.add_tag('SCTI')
        self.measurement_param = self.get_init_attr('mp', 'n.excitation', neurons)
        self.adjustment_param = 'weight_norm_factor'

        self.set_threshold(self.get_init_attr('h_sc', 0.1, neurons))
        self.adj_strength = self.get_init_attr('eta_sc', 0.001, neurons)

        self.target_clip_min = 0.5


    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            super().new_iteration(neurons)


class SORN_IP_TI(Time_Integration_Homeostasis):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.add_tag('IPTI')
        self.measurement_param = self.get_init_attr('mp', 'n.output', neurons)
        self.adjustment_param = 'TH'

        self.set_threshold(self.get_init_attr('h_ip', 0.1, neurons))
        self.adj_strength = -self.get_init_attr('eta_ip', 0.001, neurons)
        self.target_clip_min=self.get_init_attr('clip_min', -0.001, neurons)
        #target_clip_max

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            super().new_iteration(neurons)



class SORN_diffuse_IP(Time_Integration_Homeostasis):

    def set_variables(self, neurons):
        self.add_tag('diff_IP')
        super().set_variables(neurons)
        self.measurement_param = self.get_init_attr('mp', 'test_val', neurons)#output_new
        self.adjustment_param = 'TH'
        self.measure_group_sum = True

        h_dh = self.get_init_attr('h_dh', 0.1, neurons)

        #if h_dh == 'IPTI_h_ip':
        #    h_dh = np.mean((neurons['IPTI',0].min_th+neurons['IPTI',0].max_th)/2)

        self.set_threshold(h_dh)
        self.adj_strength = -self.get_init_attr('eta_dh', 0.001, neurons)


    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            super().new_iteration(neurons)



'''
def binary_activation_function(neurons):
    if neurons.rnd_act_factor != None:
        sig = neurons.rnd_act_factor * np.random.rand(neurons.size)
    else:
        sig = 0

    if hasattr(neurons, 'nox'):
        nox = neurons.nox
    else:
        nox = 0

    val = ((neurons.activity-neurons.TH-nox+sig >= 0.0) + 0.0)

    if hasattr(neurons, 'refractory_counter'):
        return val * (neurons.refractory_counter < 0.1)
    else:
        return val
    #neurons.output_new *= neurons.refractory_counter < 0.1
'''

#def identity(neurons):
#    return neurons.activity-neurons.TH

#def analog_activation_function(neurons):
#    return np.clip(np.power(neurons.activity, 2.0), 0, 1)

                #print(post_act.shape, self.pre_post_mask.shape, self.pre_post_mul.shape)
                #print(pre_act.shape, self.post_pre_mask.shape, self.post_pre_mul.shape)

                #dw = s.get_synapse_mat()

                #for t, f in self.STDP_F.items():
                #    t_start = neurons.timescale*abs(t)
                #    t_end = neurons.timescale*abs(t)+1
                #    print(t_start, t_end)
                #    if t < 0:
                #        dact = np.sum(s.dst.output_buffer[0:neurons.timescale], axis=0)
                #        sact = np.sum(s.src.output_buffer[t_start:t_end], axis=0)
                #    else:
                #        dact = np.sum(s.dst.output_buffer[t_start:t_end], axis=0)
                #        sact = np.sum(s.src.output_buffer[0:neurons.timescale], axis=0)
                #
                #    dw += f * dact[:, None] * sact[None, :]

                #print('start', neurons.tags)

                #print(list(s.dst.output_buffer_dict.keys()), list(s.src.output_buffer_dict.keys()))
                #print(len(s.dst.output_buffer_dict[neurons.timescale]), len(s.src.output_buffer_dict[neurons.timescale]))

                #dact = s.dst.get_masked_dict('output_buffer_dict', neurons.timescale)
                #sact = s.src.get_masked_dict('output_buffer_dict', neurons.timescale)

                #print(neurons.timescale)

                #print(dact.shape, sact.shape)
                #print(self.pre_post_mask.shape, self.post_pre_mask.shape)

                #masked_dact = dact[self.pre_post_mask]
                #masked_sact = sact[self.post_pre_mask]

                #print(masked_dact.shape, masked_sact.shape)

                #weighted_masked_dact = masked_dact*self.pre_post_mul
                #weighted_masked_sact = masked_sact*self.pre_post_mul

                #print(weighted_masked_dact.shape, weighted_masked_sact.shape)

                #summed_up_dact = np.sum(weighted_masked_dact, axis=0)
                #summed_up_sact = np.sum(weighted_masked_sact, axis=0)

                #print(summed_up_dact.shape, summed_up_sact.shape)

                #current_dact = dact[0]
                #current_sact = sact[0]

                #dw_pre_post = summed_up_dact[:, None] * current_sact[None, :]
                #dw_post_pre = current_dact[:, None] * summed_up_sact[None, :]

                #print(current_dact.shape, current_sact.shape)

                #o_new[:, None] * from_old[None, :] - to_old[:, None] * from_new[None, :]
                #dw = dact[:, None] * sact[None, :]

                #s.src_act_sum_old = src_out_new.copy()
            #neurons.last_output = neurons.output_buffer[0].copy()

        # for ng in self.pre_syn_groups:
        #    if not neurons.timescale in ng.buffer_timescale_sum_dict:
        #        ng.buffer_timescale_sum_dict[neurons.timescale] = np.sum(ng.output_buffer[neurons.timescale - 1:neurons.timescale * 2 - 1], axis=0)/neurons.timescale

        # for s in neurons.afferent_synapses[self.transmitter]:
        #    s.slow_add = s.W.dot(s.src.buffer_timescale_sum_dict[neurons.timescale][s.src.mask]) * self.strength

'''
    def set_variables(self, neurons):
        neurons.eta_stdp = self.get_init_attr('eta_stdp', 0.005, neurons)
        neurons.last_output = neurons.get_neuron_vec()
        for s in neurons.afferent_synapses['GLU']:
            s.src_act_sum_old = np.zeros(s.src.size)

    def new_iteration(self, neurons):
        if first_cycle_step(neurons):

            for s in neurons.afferent_synapses['GLU']:
                src_out_new = np.sum(s.src.output_buffer[0:neurons.timescale], axis=0)

                grow = s.dst.output_buffer[0][:, None] * s.src_act_sum_old[None, :]
                shrink = s.dst.last_output[:, None] * src_out_new[None, :]

                dw = neurons.eta_stdp * (grow - shrink) / (neurons.timescale*neurons.timescale)

                s.W += dw * s.enabled
                s.W[s.W < 0.0] = 0.0

                s.src_act_sum_old = src_out_new.copy()
            neurons.last_output = neurons.output_buffer[0].copy()
'''


'''
class SORN_STDP_old(Neuron_Behaviour):

    def set_variables(self, neurons):
        self.add_tag('STDP')
        neurons.eta_stdp = self.get_init_attr('eta_stdp', 0.005, neurons)
        #neurons.prune_stdp = self.get_init_attr('prune_stdp', False, neurons)

        for s in neurons.afferent_synapses['GLU']:
            s.STDP_src_lag_buffer_old = np.zeros(s.src.size)#neurons.size bug?
            s.STDP_src_lag_buffer_new = np.zeros(s.src.size)#neurons.size bug?

    def new_iteration(self, neurons):
        for s in neurons.afferent_synapses['GLU']:

            #print(s.src.size, len(s.STDP_src_lag_buffer_new), len(s.src.output_new))

            s.STDP_src_lag_buffer_new += s.src.output_buffer[0]/neurons.timescale

            #buffer previous states

            if last_cycle_step(neurons):

                from_old = s.STDP_src_lag_buffer_old#
                from_new = s.STDP_src_lag_buffer_new#s.src.x_new#

                to_old = s.dst.output_buffer[1]
                to_new = s.dst.output_buffer[0]

                #print(to_new, from_old, to_old, from_new)

                dw = neurons.eta_stdp * (to_new[:, None] * from_old[None, :] - to_old[:, None] * from_new[None, :])

                s.W += dw * s.enabled
                s.W[s.W < 0.0] = 0.0

                #if neurons.prune_stdp:
                #    s.W[s.W < 1e-10] = 0
                #    s.enabled *= (s.W > 0)

                # print('STDP',s.W.sum())

                #clear buffer
                s.STDP_src_lag_buffer_old = s.STDP_src_lag_buffer_new.copy()
                s.STDP_src_lag_buffer_new = np.zeros(s.src.size)#neurons.size bug?
'''

#class SORN_slow_input_collect(Neuron_Behaviour):#has to be executed AFTER intra, inter...behaviours
#
#    def set_variables(self, neurons):
#        self.add_tag('input_collect')
#        neurons.output_new_temp = neurons.get_neuron_vec() #required for STDP todo:remove

#    def new_iteration(self, neurons):
#        if last_cycle_step(neurons):
#            neurons.output_new_temp = neurons.activation_function(neurons)
            #reset moved to init_vars

            #neurons.test_val = (activation_function(neurons.activity, neurons) + activation_function(neurons.slow_act, neurons))/2


from PymoNNto.Exploration.Network_UI.TabBase import *

class chain_tab(TabBase):

    def __init__(self, title='Chain'):
        super().__init__(title)

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):
        if Network_UI.network['grammar_act', 0] is not None:
            self.chaintab = Network_UI.Next_Tab(self.title)

            alphabet = Network_UI.network['grammar_act'][0].alphabet
            a_list = [alphabet[i] for i in range(len(alphabet))]
            a_list[0] = '_'
            ydict = dict(enumerate(a_list))

            win = pg.GraphicsWindow()
            stringaxis = pg.AxisItem(orientation='left')
            stringaxis.setTicks([ydict.items()])

            p = Network_UI.Add_plot(axisItems={'left': stringaxis}, x_label='buffer steps')

            self.graph = pg.GraphItem()
            # p=self.Add_plot('', True)
            p.addItem(self.graph)

            source = Network_UI.network['grammar_act'][0]
            RALN = Reconstruct_Analyze_Label_Network(Network_UI.network)
            groups = Network_UI.network['prediction_source']#Network_UI.exc_group_name  # , network.NeuronGroups[1]

            def update():
                RALN.label_and_group_neurons(Network_UI.network['prediction_source', 0], [source.get_activation(char, Network_UI.network['prediction_source', 0]) for char in range(source.get_alphabet_length())], 'W', 20)
                _, self.edges = RALN.visualize_label_and_group_neurons(x_attribute_name='temporal_layer',
                                                                       y_attribute_name='class_label',
                                                                       weight_attribute_name='W',
                                                                       groups=groups,
                                                                       weight_limit=None,
                                                                       n_biggest=1,
                                                                       source=source,
                                                                       return_graph=True)  # 1/3#None#1/3#0.5

                self.nodes = RALN.get_neuron_positions(Network_UI.network['prediction_source', 0],
                                                       x_attribute_name='temporal_layer',
                                                       y_attribute_name='class_label',
                                                       y_scale=0.0005)

            self.update_btn = QPushButton('update live buffers', Network_UI.main_window)
            self.update_btn.clicked.connect(update)

            Network_UI.Next_H_Block()
            Network_UI.Add_element(self.update_btn)


            #def plot_buffer_chains(event):
            #    # self.network.NeuronGroups = [self.network.NeuronGroups[0]]
            #    # self.network.SynapseGroups = [self.network.SynapseGroups[0]]
            #    source = Network_UI.network['grammar_act', 0]
            #    RALN = Reconstruct_Analyze_Label_Network(Network_UI.network)
            #    groups = Network_UI.network[Network_UI.exc_group_name, 0]  # , network.NeuronGroups[1]
            #    RALN.label_and_group_neurons(Network_UI.network[Network_UI.exc_group_name, 0],
            #                                 [source.get_activation(char) for char in range(source.get_alphabet_length())],
            #                                 'W', 10)
            #    RALN.visualize_label_and_group_neurons(x_attribute_name='temporal_layer', y_attribute_name='class_label',
            #                                           weight_attribute_name='W', groups=groups, weight_limit=None,
            #                                           n_biggest=3, source=source)  # 1/3#None#1/3#0.5
            #
            #self.pbc_btn = QPushButton('plot buffer chains', Network_UI.main_window)
            #self.pbc_btn.clicked.connect(plot_buffer_chains)
            #Network_UI.Add_element(self.pbc_btn)
            # self.Add_Sidebar_Element(self.update_btn)


    def update(self, Network_UI):
        if Network_UI.network['grammar_act', 0] is not None and self.chaintab.isVisible() and hasattr(self, "nodes"):
            group = Network_UI.network['prediction_source', 0]

            pos = np.array(self.nodes, dtype=float)
            colors = [(act * 255, 0, 0, 50) for act in group.output]

            if Network_UI.network[Network_UI.neuron_select_group, 0] is Network_UI.network['prediction_source', 0]:
                GLU_syn = Network_UI.get_combined_syn_mats(group['GLU'])
                GLU_syn = GLU_syn[list(GLU_syn.keys())[0]]
                big_syn_indices = np.where(GLU_syn[Network_UI.neuron_select_id] > (np.max(GLU_syn[Network_UI.neuron_select_id]) * (1 / 2)))[0]
                adj = np.array([[s, Network_UI.neuron_select_id] for s in big_syn_indices])  # np.array(self.edges)##[self.neuron_select_id, 1]
                colors[Network_UI.neuron_select_id] = (0, 255, 0, 255)
                self.graph.setData(pos=pos, adj=adj, size=0.5, symbol='o', pxMode=False, symbolBrush=colors, symbolPen=None)
            else:
                self.graph.setData(pos=pos, size=0.5, symbol='o', pxMode=False, symbolBrush=colors, symbolPen=None)

            self.graph.update()

            #selected_GLU_syn = GLU_syn[Network_UI.neuron_select_id]

            #GABA_syn = Network_UI.network[Network_UI.neuron_select_group, 0]['GABA']
            #if len(GABA_syn) > 0:
            #    GABA_syn = Network_UI.get_combined_syn_mats(GABA_syn)
            #    GABA_syn = GABA_syn[list(GABA_syn.keys())[0]]
            #    selected_GABA_syn = GABA_syn[Network_UI.neuron_select_id]
            #else:
            #    GABA_syn = None



from PymoNNto.Exploration.StorageManager.StorageManager import *
import sys

evolution_genome = None
default_genome = {}

def set_evolution_genome_from_string(gene_str):
    #print('gene_str', gene_str)
    gene_dict = {}
    partial_str = gene_str.split('#')
    for key_value_str in partial_str:
        kv_partial_str = key_value_str.split('@')
        if len(kv_partial_str) == 2:
            if kv_partial_str[0] != 'score':
                gene_dict[kv_partial_str[0]] = kv_partial_str[1] #keys and values are strings!!! have to be casted to default type by get_gene!

    set_genome(gene_dict)

def set_genome(genome):
    #print('set genome to', genome)
    if 'evo_name' in genome:
        global evolution_genome
        evolution_genome = genome
    else:
        print('no evo_name key in evolution for', genome)

def get_genome():
    if evolution_genome is None:
        load_genome()
    return evolution_genome

def cast_to_default(str_value, default):
    if isinstance(default, float) or isinstance(default, int):
        return float(str_value)
    else:
        return str_value

def get_gene(key, default):

    if evolution_genome is None:
        load_genome()

    default_genome[key] = default
    if evolution_genome is not None and key in evolution_genome:
        return cast_to_default(evolution_genome[key], default)
    else:
        return default

def return_default_genome():
    return #todo:implement

def gene(key, default):
    return get_gene(key, default)

def get_gene_id(gene):
    id = ''
    for key, value in gene.items():
        id += '#'+key+'@'+str(value)
    return id+'#'

def get_gene_file(gene):
    return gene['evo_name']+' gen'+str(gene['gen']) + ' id' + str(gene['id'])

def set_score(score):
    global evolution_genome

    if evolution_genome is not None:
        if 'score' in evolution_genome:#only when setscore is called multiple times on accident...
            evolution_genome.pop('score')
        sm = StorageManager(main_folder_name=get_gene('evo_name', None), folder_name=get_gene_file(evolution_genome), print_msg=False, add_new_when_exists=False)
        evolution_genome['score'] = score
        sm.save_param_dict(evolution_genome)
        print('evolution score set to #'+score+'#')
    else:
        print('no genome found for score')

    # reset and reload when next get_gene is called

    evolution_genome = None


def load_genome():
    for arg in sys.argv:
        if 'genome=' in arg:
            set_evolution_genome_from_string(arg[7:])



from PymoNNto.Exploration.Network_UI.TabBase import *

class stdp_buffer_tab(TabBase):

    def __init__(self, title='STDP'):
        super().__init__(title)

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):
        self.stdp_tab = Network_UI.Next_Tab(self.title)#, create_h_block=False

        Network_UI.Add_element(QLabel('Select syn'), stretch=0)
        self.select_syn_box = QComboBox()
        self.select_syn_box.addItems(Network_UI.transmitters)
        self.select_syn_box.setCurrentIndex(0)
        Network_UI.Add_element(self.select_syn_box, stretch=4)

        Network_UI.Next_H_Block(stretch=0)
        self.min_label=QLabel('???')
        Network_UI.Add_element(self.min_label, stretch=4)

        Network_UI.Next_H_Block(stretch=0)
        self.max_label=QLabel('???')
        Network_UI.Add_element(self.max_label, stretch=4)

        Network_UI.Next_H_Block(stretch=10)

        grid_base_widget = QWidget()
        Network_UI.Add_element(grid_base_widget)


        grid_layout = QGridLayout(grid_base_widget)
        #Network_UI.visualization_layout.addLayout(grid_layout)

        self.syn_canvas = pg.GraphicsLayoutWidget()

        self.syn_canvas.setBackground((255, 255, 255))
        self.syn_plot = self.syn_canvas.addPlot(row=0, col=0)
        self.syn_plot.setToolTip('Synaptic change from STDP (gray=no change, black=shrink, white=grow)')
        self.syn_plot.setLabels(title='Delta W')
        #self.syn_plot.hideAxis('left')
        #self.syn_plot.hideAxis('bottom')
        self.syn_img = pg.ImageItem(np.random.rand(291, 291, 3))
        self.syn_plot.addItem(self.syn_img)
        grid_layout.addWidget(self.syn_canvas, 0, 0)

        def update():
            base_rect = self.syn_plot.viewRect()
            self.pre_plot.setRange(QtCore.QRectF(base_rect.x(), 0, base_rect.width(), self.pre_length), padding=0)  #
            self.post_plot.setRange(QtCore.QRectF(0, base_rect.y(), self.post_length, base_rect.height()), padding=0)  #
            self.syn_plot.disableAutoRange()
            self.pre_plot.disableAutoRange()
            self.post_plot.disableAutoRange()

        old_wheel_event=self.syn_canvas.wheelEvent
        def wheel(event):
            old_wheel_event(event)
            update()
        self.syn_canvas.wheelEvent=wheel

        old_move_event=self.syn_canvas.mouseMoveEvent
        def move(event):
            old_move_event(event)
            update()
        self.syn_canvas.mouseMoveEvent=move

        self.pre_canvas = pg.GraphicsLayoutWidget()
        self.pre_canvas.setBackground((255, 255, 255))
        self.pre_plot = self.pre_canvas.addPlot(row=0, col=0)
        self.pre_plot.setToolTip('Pre Synaptic Activity')
        self.pre_plot.setLabels(title='Pre Act Buf')
        #self.pre_plot.hideAxis('left')
        #self.pre_plot.hideAxis('bottom')
        self.pre_img = pg.ImageItem(np.random.rand(291, 291, 3))
        self.pre_plot.addItem(self.pre_img)
        grid_layout.addWidget(self.pre_canvas, 1, 0)

        self.post_canvas = pg.GraphicsLayoutWidget()

        self.post_canvas.setBackground((255, 255, 255))
        self.post_plot = self.post_canvas.addPlot(row=0, col=0)
        self.post_plot.setToolTip('Post Synaptic Activity')
        self.post_plot.setLabels(title='Post Act Buf')
        #self.post_plot.hideAxis('left')
        #self.post_plot.hideAxis('bottom')
        self.post_img = pg.ImageItem(np.random.rand(291, 291, 3))
        self.post_plot.addItem(self.post_img)
        grid_layout.addWidget(self.post_canvas, 0, 1)

        grid_layout.setColumnStretch(0, 10)
        grid_layout.setColumnStretch(1, 1)
        grid_layout.setRowStretch(0, 6)
        grid_layout.setRowStretch(1, 1)

        Network_UI.Next_H_Block(stretch=0)

        qc = QCheckBox('Axis')
        def change(event):
            axis_enabeled = qc.isChecked()
            self.syn_plot.showAxis('left', axis_enabeled)
            self.syn_plot.showAxis('bottom', axis_enabeled)
            self.pre_plot.showAxis('left',axis_enabeled )
            self.pre_plot.showAxis('bottom', axis_enabeled)
            self.post_plot.showAxis('left', axis_enabeled)
            self.post_plot.showAxis('bottom', axis_enabeled)
        qc.setChecked(False)
        change(None)
        qc.stateChanged.connect(change)
        Network_UI.Add_element(qc)

        self.sensitivity_slider = QSlider(1)
        self.sensitivity_slider.setMinimum(0)
        self.sensitivity_slider.setMaximum(100)
        self.sensitivity_slider.setSliderPosition(50)
        Network_UI.Add_element(self.sensitivity_slider)

        self.current_selector_group=None
        self.current_selector_index_dict={}

    def get_shift_and_zoom(self, transform):
        zero = QtCore.QPointF(0, 0)
        p = QtCore.QPointF(1, 1)
        zero_t = transform.map(zero)
        p_t = transform.map(p)
        x_shift = zero_t.x()
        y_shift = zero_t.y()
        x_zoom = p_t.x() - x_shift
        y_zoom = p_t.y() - y_shift
        return x_shift, y_shift, x_zoom, y_zoom

    def update(self, Network_UI):
        if self.stdp_tab.isVisible():
            if len(Network_UI.network[Network_UI.neuron_select_group]) > 0:
                group = Network_UI.network[Network_UI.neuron_select_group, 0]
                synapse_groups = group.afferent_synapses['All']
                #print(synapse_groups)

                if self.current_selector_group != group:
                    self.current_selector_group = group
                    self.select_syn_box.clear()
                    self.current_selector_index_dict.clear()
                    self.current_selector_index_dict[-1]=-1
                    ct=0
                    for i,s in enumerate(synapse_groups):
                        if (type(s.dst.mask) == np.ndarray and s.dst.mask[Network_UI.neuron_select_id]) or (type(s.dst.mask) is bool and s.dst.mask==True):
                            self.current_selector_index_dict[ct]=i
                            self.select_syn_box.addItem(' '.join(s.tags))
                            ct += 1

                    #self.select_syn_box.addItems(Network_UI.transmitters)

                current_select_index = self.current_selector_index_dict[self.select_syn_box.currentIndex()]

                if current_select_index >= 0:
                    syn = synapse_groups[current_select_index]

                    post_act = group.get_buffer(syn.dst, 'output', group.timescale)#s.dst.get_masked_dict('output_buffer_dict', neurons.timescale)
                    pre_act = group.get_buffer(syn.src, 'output', group.timescale)#s.src.get_masked_dict('output_buffer_dict', neurons.timescale)

                    self.post_length = len(post_act)
                    self.pre_length = len(pre_act)

                    sensitivity = self.sensitivity_slider.sliderPosition()/100.0*0.0001

                    self.pre_img.setImage(np.rot90(pre_act, 3), levels=(0, 1))
                    self.post_img.setImage(np.rot90(post_act.transpose(), 3), levels=(0, 1))

                    if hasattr(syn, 'dw'):
                        self.min_label.setText('min dW: ' + str(np.min(syn.dw)))
                        self.max_label.setText('max dW: ' + str(np.max(syn.dw)))
                        self.syn_img.setImage(np.rot90(syn.dw, 3), levels=(-sensitivity, +sensitivity))
                    else:
                        self.syn_img.clear()

            else:
                self.syn_img.clear()
                self.pre_img.clear()
                self.post_img.clear()

            '''
                    gap = 5
    
                    dwh = syn.dw.shape[0]
                    dww = syn.dw.shape[1]
    
                    syn.dw = np.concatenate([syn.dw, np.ones((syn.dw.shape[0], gap + len(post_act))) * -1], axis=1)
                    syn.dw = np.concatenate([syn.dw, np.ones((gap + len(pre_act), syn.dw.shape[1])) * -1], axis=0)
    
                    syn.dw[0:len(post_act[0]), dww + gap:dww + gap + len(post_act)] = post_act.transpose()
    
                    syn.dw[dwh + gap:dwh + gap + len(pre_act), 0:len(pre_act[0])] = pre_act
            '''







from PymoNNto.NetworkBehaviour.Input.Activator import *
from PymoNNto.NetworkBehaviour.Logic.SORN.SORN_advanced_buffer import *



##########################################################################
#Generate output with k winner takes all algorithm
##########################################################################
class SORN_generate_output_K_WTA(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('K_WTA')

        neurons.output = neurons.get_neuron_vec()

        self.K = self.get_init_attr('K', 10, neurons)

        if self.K < 1:
            self.K = int(neurons.size * self.K)


    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            ind = np.argpartition(neurons.activity, -self.K)[-self.K:]
            neurons.output.fill(0)
            neurons.output[ind] = 1


class SORN_generate_output_K_WTA_partitioned(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('K_WTA_partitioned')

        self.filter_temporal_output = self.get_init_attr('filter_temporal_output', False, neurons)



        self.K = self.get_init_attr('K', 0.1, neurons)#only accepts values between 0 and 1

        partition_size = self.get_init_attr('partition_size', 7, neurons)
        self.partitioned_ng=neurons.partition_size(partition_size)

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):

            #for syn in neurons.afferent_synapses['GLU']:
            #    ng=syn.dst
            for ng in self.partitioned_ng:#
                K = ng.size * self.K
                #for non integer K
                K_floor = int(np.floor(K))
                if np.random.rand() < (K-K_floor):
                    K = K_floor+1
                else:
                    K = K_floor

                ng.output *= 0

                if K>0:
                    act = ng.activity.copy()

                    if self.filter_temporal_output:
                        act = ng.activity*ng.output#-(s.dst.output*-10000)

                    ind = np.argpartition(act, -K)[-K:]
                    act_mat = np.zeros(ng.size)
                    act_mat[ind] = 1

                    ng.output += act_mat

                    #s.dst.output = s.dst.output*0.0+act_mat
                    #
                    #s.dst.output[ind] += 1


                #s.dst._temp_act_sum += np.mean(s.src.output)


class SORN_WTA_fast_syn(SORN_signal_propagation_base):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.add_tag('fast_' + self.transmitter)
        self.get_init_attr('so', None, neurons)#just for error suppression

    def new_iteration(self, neurons):
        if last_cycle_step(neurons) and self.strength != 0:
            for s in neurons.afferent_synapses[self.transmitter]:
                s.fast_add = s.W.dot(s.src.output) * self.strength# / neurons.timescale

                s.dst.activity += s.fast_add
                if self.strength > 0:
                    s.dst.excitation += s.fast_add
                else:
                    s.dst.inhibition += s.fast_add

class SORN_WTA_iSTDP(Behaviour):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.add_tag('WTA iSTDP')
        self.eta_iSTDP = self.get_init_attr('eta_iSTDP', 0.1, neurons)

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            for s in neurons.afferent_synapses['GABA']:
                add = s.dst.activity[:, None] * s.src.activity[None, :] * self.eta_iSTDP * s.enabled
                #print(add.shape)
                s.W += add


class SORN_IP_TI_WTA(Time_Integration_Homeostasis):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.add_tag('IP_TI_WTA')
        self.measurement_param = self.get_init_attr('mp', 'n.output', neurons)
        self.adjustment_param = 'exhaustion_value'

        self.set_threshold(self.get_init_attr('h_ip', 0.1, neurons))
        self.adj_strength = -self.get_init_attr('eta_ip', 0.001, neurons)
        self.target_clip_min=self.get_init_attr('clip_min', -0.001, neurons)

        #target_clip_max

        neurons.exhaustion_value = 0

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            super().new_iteration(neurons)
            neurons.activity -= neurons.exhaustion_value


class SORN_IP_WTA(Instant_Homeostasis):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.add_tag('IP_WTA')
        self.measurement_param = self.get_init_attr('mp', 'n.output', neurons)
        self.adjustment_param = 'exhaustion_value'

        self.set_threshold(self.get_init_attr('h_ip', 0.1, neurons))
        self.adj_strength = -self.get_init_attr('eta_ip', 0.001, neurons)

        #target_clip_max

        neurons.exhaustion_value = 0

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            super().new_iteration(neurons)

            neurons.exhaustion_value = neurons.exhaustion_value-np.mean(neurons.exhaustion_value)

            #neurons.activity -= neurons.exhaustion_value

class SORN_IP_WTA_apply(Behaviour):

    def set_variables(self, neurons):
        super().set_variables(neurons)
        self.add_tag('SORN_IP_WTA_apply')

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            neurons.activity -= neurons.exhaustion_value

'''
class SORN_Neuron_Exhaustion(Neuron_Behaviour):

    def set_variables(self, neurons):
        self.add_tag('exhaustion')
        neurons.exhaustion_value = neurons.get_neuron_vec()
        self.decay_factor = self.get_init_attr('decay_factor', 0.9, neurons)
        self.strength = self.get_init_attr('strength', 0.1, neurons)

    def new_iteration(self, neurons):
        if last_cycle_step(neurons):
            neurons.exhaustion_value *= self.decay_factor
            neurons.exhaustion_value += neurons.output

            neurons.activity -= neurons.exhaustion_value * self.strength
'''


def max_neuron_frequency():
    #if pc ...
    return 200.0

def min_spike_interval():
    return 1000.0/max_neuron_frequency()

def GammaBlock_to_ms():
    return 25.0

def GammaBlock_to_Hz(x):
    return x*(1000/(GammaBlock_to_ms()/min_spike_interval()))


from PymoNNto.Exploration.Network_UI.TabBase import *

from PymoNNto.Exploration.Visualization.Reconstruct_Analyze_Label.Reconstruct_Analyze_Label import *

class reconstruction_tab(TabBase):

    def __init__(self, title='recon'):
        super().__init__(title)

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):
        if Network_UI.network['grammar_act', 0] is not None:
            self.reconstruction_tab = Network_UI.Next_Tab(self.title)

            self.grid = QGridLayout()
            self.grid.setAlignment(Qt.AlignLeft)
            Network_UI.current_H_block.addLayout(self.grid)
            Network_UI.current_H_block.setAlignment(Qt.AlignTop)

            source = Network_UI.network['grammar_act', 0]
            self.labels = []

            for y, char in enumerate(source.alphabet):
                self.labels.append([])
                for timestep in range(11):
                    label = QLabel('<font color='+('#%02x%02x%02x'% (timestep*25,timestep*25,timestep*25)).upper()+'>'+char.replace(' ','_')+'</font>')
                    label.char=char
                    self.labels[-1].append(label)
                    #label.setFont(QFont())
                    self.grid.addWidget(label, y, timestep)

            self.img = Network_UI.Add_Image_Item(False, False, title=' neuron rec')

            self.recon_text_label = QLabel()
            Network_UI.Add_element(self.recon_text_label)

            #Network_UI.Next_H_Block()

            self.net_grid = QGridLayout()
            self.net_grid.setAlignment(Qt.AlignLeft)
            Network_UI.current_H_block.addLayout(self.net_grid)
            Network_UI.current_H_block.setAlignment(Qt.AlignTop)

            self.net_labels = []

            for y, char in enumerate(source.alphabet):
                self.net_labels.append([])
                for timestep in range(11):
                    net_label = QLabel('<font color='+('#%02x%02x%02x'% (timestep*25,timestep*25,timestep*25)).upper()+'>'+char.replace(' ','_')+'</font>')
                    net_label.char = char
                    self.net_labels[-1].append(net_label)
                    #label.setFont(QFont())
                    self.net_grid.addWidget(net_label, y, timestep)

            self.net_img = Network_UI.Add_Image_Item(False, False, title='net rec')

            self.net_recon_text_label = QLabel()
            Network_UI.Add_element(self.net_recon_text_label)






    def update(self, Network_UI):
        if Network_UI.network['grammar_act', 0] is not None and self.reconstruction_tab.isVisible():
            group=Network_UI.network[Network_UI.neuron_select_group, 0]

            RALN = Reconstruct_Analyze_Label_Network(Network_UI.network)
            RALN.zero_recon()
            group.recon[Network_UI.neuron_select_id] = 1
            RALN.propagation('W', 10, 'backward', 'forget', 'all', temporal_recon_groups=Network_UI.network['prediction_source'], exponent=4, normalize=True, filter_weakest_percent=40.0)  # forget

            source = Network_UI.network['grammar_act', 0]

            for ng in Network_UI.network['prediction_source']:
                baseline = ng.Input_Weights.transpose().dot(ng.temporal_recon[-1])
                baseline = baseline/np.sum(baseline)

                if ng == group:#clicked group?
                    temp = ng.get_neuron_vec()
                    temp[Network_UI.neuron_select_id] = 1.0
                    ng.temporal_recon.insert(0, temp)
                else:
                    ng.temporal_recon.append(ng.get_neuron_vec())

                text = ''
                res = []

                for r in ng.temporal_recon:
                    char_vec = ng.Input_Weights.transpose().dot(r)
                    s = np.sum(char_vec)
                    if s > 0:
                        char_vec = char_vec/s
                        res.insert(0, char_vec-baseline)
                        text = source.index_to_char(np.argmax(char_vec)) + text
                    else:
                        res.insert(0, np.zeros(ng.Input_Weights.shape[1]))
                        text = '#' + text

                text = list(text)
                start = text[0]
                for i, c in enumerate(text):
                    if text[i] == start:
                        text[i] = '#'
                    else:
                        break
                text = ''.join(text)

                res = np.array(res)
                res = res-np.min(res)
                self.img.setImage(np.fliplr(res.copy()))
                #res = np.array(res*res)

                for y in range(res.shape[0]):
                    for x in range(res.shape[1]):
                        m = (np.max(res)+np.max(res[y, :]))/2.0
                        if m == 0:
                            m = 1.0
                        val = 255-np.clip(int((res[y, x]-(np.mean(res[y, :])/2.0))/m*255.0), 0, None)
                        self.labels[x][y].setText('<font color='+('#%02x%02x%02x' % (val, val, val)).upper()+'>'+self.labels[x][y].char.replace(' ','_')+'</font>')

                self.recon_text_label.setText(text)

        ################################################################################################################
        ################################################################################################################
        ################################################################################################################

            RALN = Reconstruct_Analyze_Label_Network(Network_UI.network)
            RALN.zero_recon()
            for ng in Network_UI.network.NeuronGroups:
                ng.recon = ng.output.copy()
            RALN.propagation('W', 10, 'backward', 'forget', 'all', temporal_recon_groups=Network_UI.network['prediction_source'], exponent=4, normalize=True, filter_weakest_percent=40.0)  # forget

            source = Network_UI.network['grammar_act', 0]

            for ng in Network_UI.network['prediction_source']:
                baseline = ng.Input_Weights.transpose().dot(ng.temporal_recon[-1])
                baseline = baseline / np.sum(baseline)

                ng.temporal_recon.append(ng.get_neuron_vec())

                text = ''
                res = []

                for r in ng.temporal_recon:
                    char_vec = ng.Input_Weights.transpose().dot(r)
                    s = np.sum(char_vec)
                    if s > 0:
                        char_vec = char_vec / s
                        res.insert(0, char_vec - baseline)
                        text = source.index_to_char(np.argmax(char_vec)) + text
                    else:
                        res.insert(0, np.zeros(ng.Input_Weights.shape[1]))
                        text = '#' + text

                text = list(text)
                start = text[0]
                for i, c in enumerate(text):
                    if text[i] == start:
                        text[i] = '#'
                    else:
                        break
                text = ''.join(text)

                res = np.array(res)
                res = res - np.min(res)
                self.net_img.setImage(np.fliplr(res.copy()))
                # res = np.array(res*res)

                for y in range(res.shape[0]):
                    for x in range(res.shape[1]):
                        m = (np.max(res) + np.max(res[y, :])) / 2.0
                        if m == 0:
                            m = 1.0

                        val = 255 - np.clip(int((res[y, x] - (np.mean(res[y, :]) / 2.0)) / m * 255.0), 0, None)
                        self.net_labels[x][y].setText('<font color=' + ('#%02x%02x%02x' % (val, val, val)).upper() + '>' + self.net_labels[x][y].char.replace(' ','_') + '</font>')

                self.net_recon_text_label.setText(text)



from PymoNNto.Exploration.UI_Base import *
from PymoNNto.NetworkBehaviour.Recorder.Recorder import *

class Network_UI(UI_Base):

    def __init__(self, network, modules=[], label='SORN UI', group_tags=[], transmitters=[], storage_manager=None, group_display_count=None, reduced_layout=False):

        network.clear_recorder()

        #network.simulate_iteration()
        self.recording = False
        if self.recording:
            label += ' rec.'

        for ng in network.NeuronGroups:
            if not hasattr(ng, 'color'):
                ng.color = (0, 0, 255, 255)

        if group_tags==[]:
            for ng in network.NeuronGroups:
                if ng.tags[0] not in group_tags:
                    group_tags.append(ng.tags[0])

        if transmitters==[]:
            for sg in network.SynapseGroups:
                if sg.tags[0] not in transmitters:
                    transmitters.append(sg.tags[0])

        self.reduced_layout=reduced_layout

        #for group in network[inh_group_name]:
        #    network.add_behaviours_to_neuron_group({10000: Recorder(['np.mean(n.output)',
        #                           'np.mean(n.TH)',
        #                           'n.TH',
        #                           'n.excitation',
        #                           'n.inhibition',
        #                           'n.input_act',
        #                           'n.refractory_counter',
        #                           '[np.sum(s.slow_add) for s in n.afferent_synapses.get("All")]',
        #                           '[np.sum(s.fast_add) for s in n.afferent_synapses.get("All")]'], tag='UI_rec')}, group)

        super().__init__(network, label=label)

        self.main_window.keyPressEvent = self.keyPressEvent

        self.group_display_count=group_display_count

        #self.exc_group_name = exc_group_name
        #self.inh_group_name = inh_group_name
        self.group_tags=group_tags
        self.transmitters=transmitters
        self.pause = False
        self.update_without_state_change = False
        self.storage_manager = storage_manager

        self.neuron_select_x = 0
        self.neuron_select_y = 0
        self.neuron_select_id = 0
        self.neuron_select_group = group_tags[0]#exc_group_name
        self.neuron_visible_groups = []
        #self.ts_group = 0
        #self.x_steps = 500
        self.group_sliders=[]
        self.neuron_select_color=(0,255,0,255)

        self.modules = modules

        for module in self.modules:
            print('Initialize:', module)
            module.initialize(self)

        for group_tag in group_tags:
            for group in network[group_tag]:

                group._rec_dict={}

                #rec = Recorder([], tag='UI_rec')
                #network.add_behaviours_to_neuron_group({10000: rec}, group)

                for module in self.modules:
                    module.add_recorder_variables(group, self)

        self.init_recoders()

        timer = QtCore.QTimer(self)
        timer.timeout.connect(self.on_timer)
        timer.start(40)

    def add_recording_variable(self, group, var, timesteps):

        old_ts=0
        if var in group._rec_dict:
            old_ts=group._rec_dict[var]

        group._rec_dict[var] = max(timesteps,old_ts)
        #recorder.add_varable('n.output')

    def init_recoders(self):
        for group_tag in self.group_tags:
            for group in self.network[group_tag]:

                rec_time_dict={}
                for variable in group._rec_dict:
                    rec_length=group._rec_dict[variable]
                    if rec_length not in rec_time_dict:
                        rec_time_dict[rec_length]=[]
                    rec_time_dict[rec_length].append(variable)

                for rec_length in rec_time_dict:
                    rec = Recorder(rec_time_dict[rec_length] + ['n.iteration'], tag='UI_rec,rec_' + str(rec_length), max_length=rec_length)
                    self.network.add_behaviours_to_object({10000+rec_length: rec}, group)

    #def rec(self, neuron_group, rec_length=-1):
    #    return neuron_group[self.rec_tag(rec_length),0]

    #def rec_tag(self, rec_length=-1):
    #    if rec_length==-1:
    #        rec_length = self.default_rec_recording_length
    #    return 'rec_'+str(rec_length)

    def static_update_func(self, event=None):
        if self.pause:
            self.update_without_state_change=True


    def record_image(self, img, key):
        if self.storage_manager is not None and self.recording:
            image = upscale(img*255, 10)
            self.storage_manager.save_frame(image, key)

    def get_selected_neuron_subgroup(self):
        syn_sgs = self.get_selected_synapses()
        if len(syn_sgs) > 0:
            return syn_sgs[0]
        else:
            return None

    def get_selected_synapses(self):
        result = []
        if len(self.network[self.neuron_select_group]) > 0:
            group = self.network[self.neuron_select_group, 0]
            synapse_groups = group.afferent_synapses['All']
            for i, s in enumerate(synapse_groups):
                if (type(s.dst.mask) == np.ndarray and s.dst.mask[self.neuron_select_id]) or (type(s.dst.mask) is bool and s.dst.mask == True):
                    result.append(synapse_groups[i].dst)
        return result

    def get_combined_syn_mats(self, synapses, neuron_id=None, attr='W'):
        results = {}
        shapes = {}
        for s in synapses:
            base_src = s.src.group_without_subGroup()
            base_dst = s.dst.group_without_subGroup()
            key = ','.join(s.tags)
            if not key in results:
                results[key] = np.zeros((base_dst.size, base_src.size))
                shapes[key] = (base_src.height, base_src.width)
            try:
                syn_mat=eval('s.'+attr)
                if base_src == s.src and base_dst == s.dst:
                    results[key] += syn_mat#.copy()
                else:
                    mat_mask = s.dst.mask[:, None] * s.src.mask[None, :]
                    results[key][mat_mask] += np.array(syn_mat).flatten() #np.array required if syn_mat is bool (enabled)
            except:
                print(attr, "cannot be evaluated")

        if neuron_id is not None:
            for key in results:
                results[key] = results[key][neuron_id].reshape(shapes[key])

        return results


    def on_timer(self):

        if not self.pause or self.step or self.update_without_state_change:
            self.step = False
            if not self.update_without_state_change:
                for i in range(1):
                    self.network.simulate_iteration()

            self.it = self.network.iteration

            for module in self.modules:
                module.update(self)

            #for rec in self.network['UI_rec']:
            #    rec.cut_length(self.default_recorder_length)

            self.update_without_state_change = False


    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_W:
            indx = self.tabs.currentIndex()
            if indx >= 0:
                widget=self.tabs.currentWidget()
                self.tabs.removeTab(indx)
                widget.setParent(None)
                widget.show()







def get_color(type_index, layer):
    dim_value = max(layer * 0.9, 1.0)

    if type_index == 0:
        return (0.0, 0.0, 255.0 / dim_value, 255.0)
    if type_index == 1:
        return (255.0 / dim_value, 0.0, 0.0, 255.0)
    if type_index == 2:
        return (255.0 / dim_value, 150.0 / dim_value, 0.0, 255.0)
    if type_index == 3:
        return (255.0 / dim_value, 80.0 / dim_value, 0.0, 255.0)
    if type_index == 4:
        return (255.0 / dim_value, 0.0 , 150.0/ dim_value, 255.0)


'''
def get_color(type_index, layer):
    dim_value = max(layer * 1.0, 1.0)

    if type_index == 0:
        return (0.0, 0.0, 255.0 / dim_value, 255.0)
    if type_index == 1:
        return (255.0 / dim_value, 0.0, 0.0, 255.0)
    if type_index == 2:
        return (255.0 / dim_value, 150.0 / dim_value, 0.0, 255.0)
    if type_index == 3:
        return (255.0 / dim_value, 80.0 / dim_value, 0.0, 255.0)
    if type_index == 4:
        return (255.0 / dim_value, 0.0 , 150.0/ dim_value, 255.0)
'''

########################################################### Exception handling



from io import StringIO
import traceback
from PyQt5 import QtCore
from PyQt5.QtWidgets import *
#import time

def excepthook(excType, excValue, tracebackobj):
    """
    Global function to catch unhandled exceptions.

    @param excType exception type
    @param excValue exception value
    @param tracebackobj traceback object
    """
    separator = '-' * 80
    logFile = "simple.log"
    notice = \
        """An unhandled exception occurred. Please report the problem\n""" \
        """using the error reporting dialog or via email to <%s>.\n""" \
        """A log has been written to "%s".\n\nError information:\n""" % \
        ("yourmail at server.com", "")
    versionInfo = "0.0.1"
    timeString = time.strftime("%Y-%m-%d, %H:%M:%S")

    tbinfofile = StringIO()
    traceback.print_tb(tracebackobj, None, tbinfofile)
    tbinfofile.seek(0)
    tbinfo = tbinfofile.read()
    errmsg = '%s: \n%s' % (str(excType), str(excValue))
    sections = [separator, separator, errmsg, separator, tbinfo]
    msg = '\n'.join(sections)
    #try:
    #    f = open(logFile, "w")
    #    f.write(msg)
    #    f.write(versionInfo)
    #    f.close()
    #except IOError:
    #    pass
    #errorbox = QMessageBox()
    #errorbox.setText(str(notice) + str(msg) + str(versionInfo))
    #errorbox.exec_()

sys.excepthook = excepthook




#from Exploration.Visualization.Visualization_Helper import *

#sys._excepthook = sys.excepthook
#def exception_hook(exctype, value, traceback):
#    print(exctype, value, traceback)
#    sys._excepthook(exctype, value, traceback)
#    sys.exit(1)
#sys.excepthook = exception_hook





# def record_frame(self, item=None, key='frame', width=100):
#    if self.storage_manager is not None and item is not None:
#        exporter = pg.exporters.ImageExporter(item)
#        #exporter.parameters()['width'] = width  # (note this also affects height parameter)

#        exporter.params.param('width').setValue(150, blockSignal=exporter.widthChanged)
#        exporter.params.param('height').setValue(120, blockSignal=exporter.heightChanged)

#        next=self.storage_manager.get_next_frame_name(key)
#        exporter.export(next)


#screen = QApplication.primaryScreen()
#p = screen.grabWindow(self.main_window.winId())
#p.save('test{}.png'.format(self.network.iteration), 'png')

'''
    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Left:
            self.ny-=1
        if event.key() == QtCore.Qt.Key_Right:
            self.ny+=1
        if event.key() == QtCore.Qt.Key_Up:
            self.nx+=1
        if event.key() == QtCore.Qt.Key_Down:
            self.nx-=1
        self.nx = np.clip(self.nx, 0, self.ng_width - 1)
        self.ny = np.clip(self.ny, 0, self.ng_height - 1)
        self.n_id=self.ny*self.ng_width+self.nx
'''


# if not self.update_without_state_change:
#    self.avg_big_synapses_data.append(np.average(np.sum(GLU_syn > (np.max(GLU_syn, axis=1) * (1 / 2))[:, None], axis=0)))
#    self.neuron_big_synapses_data.append(np.sum(GLU_syn[self.neuron_select_id] > (np.max(GLU_syn[self.neuron_select_id]) * (1 / 2))))

# self.avg_big_synapses_curve.setData(np.arange(it - len(self.avg_big_synapses_data), it), self.avg_big_synapses_data)
# self.neuron_big_synapses_curve.setData(np.arange(it-len(self.neuron_big_synapses_data), it), self.neuron_big_synapses_data)


'''
GLU_syn = self.get_combined_syn_mat(self.network[self.neuron_select_group, ts_group]['GLU'])[0]

selected_GLU_syn = GLU_syn[self.neuron_select_id]

GABA_syn = self.network[self.neuron_select_group, ts_group]['GABA']
if len(GABA_syn) > 0:
    GABA_syn = self.get_combined_syn_mat(GABA_syn)[0]
    selected_GABA_syn = GABA_syn[self.neuron_select_id]
else:
    GABA_syn = None





asfdsf



#exc_shape = (self.network[self.exc_group_name, ts_group].height, self.network[self.exc_group_name, ts_group].width)
w_img = np.reshape(selected_GLU_syn, exc_shape)
self.weight_GLU_items[0].setImage(w_img)

if GABA_syn is not None and selected_GABA_syn is not None:
    #inh_shape = (self.network[self.inh_group_name, ts_group].height, self.network[self.inh_group_name, ts_group].width)
    w_img = np.reshape(selected_GABA_syn, inh_shape)
    self.weight_GABA_items[0].setImage(w_img)
else:
    self.weight_GABA_items[0].clear()
'''

'''
        self.graph = pg.GraphItem()
        p=self.Add_plot('', True)
        p.addItem(self.graph)
        #self.Add_Sidebar_Element(self.graph)
        #self.inp_text_label.setText('test')

        # Define positions of nodes
        pos = np.array([
            [0, 0],
            [10, 0],
            [0, 10],
            [10, 10],
            [5, 5],
            [15, 5]
        ], dtype=float)

        # Define the set of connections in the graph
        adj = np.array([
            [0, 1],
            [1, 3],
            [3, 2],
            [2, 0],
            [1, 5],
            [3, 5],
        ])

        # Define the symbol to use for each node (this is optional)
        symbols = ['o', 'o', 'o', 'o', 'o', 'o']

        # Define the line style for each connection (this is optional)
        lines = np.array([
            (255, 255, 255, 255, 1),
            (255, 255, 255, 255, 2),
            (255, 255, 255, 255, 3),
            (255, 255, 255, 255, 2),
            (255, 255, 255, 255, 1),
            (255, 255, 255, 255, 4),
        ], dtype=[('red', np.ubyte), ('green', np.ubyte), ('blue', np.ubyte), ('alpha', np.ubyte), ('width', float)])

        # Define text to show next to each symbol
        texts = ["Point %d" % i for i in range(6)]

        # Update the graph
        self.graph.setData(pos=pos, adj=adj, pen=lines, size=1, symbol=symbols, pxMode=False, text=texts)
        #https://stackoverflow.com/questions/46868432/pyqtgraph-change-color-of-node-and-its-edges-on-click
'''

# def show_info(event):
#    self.info_window.show()

# self.info_btn = QPushButton('info', self.main_window)
# self.info_btn.clicked.connect(show_info)
# self.Add_Sidebar_Element(self.info_btn)
'''
self.gaba=None
def click2(event):
    if self.gaba is not None:
        print('a')
        self.network.NeuronGroups[0].afferent_synapses['GABA'] += self.gaba
        self.gaba=None
    else:
        print('b')
        self.gaba=self.network.NeuronGroups[0].afferent_synapses['GABA']
        self.network.NeuronGroups[0].afferent_synapses['GABA']=[x for x in self.network.NeuronGroups[0].afferent_synapses['GABA'] if x not in self.gaba]
self.btn2 = QPushButton('inhibition on/off', self.main_window)
self.btn2.clicked.connect(click2)
self.Add_Sidebar_Element(self.btn2)
'''

# canvas = pg.GraphicsLayoutWidget()
# canvas.setBackground((255, 255, 255))
# self.Add_Sidebar_Element(canvas)
# self.plot_main = canvas.addPlot(row=0, col=0)
# self.plot_main.hideAxis('left')
# self.plot_main.hideAxis('bottom')
# self.main_item = pg.ImageItem(np.random.rand(291, 291, 3))
# self.plot_main.addItem(self.main_item)

# p.addItem(pg.TextItem(text=, color=(0, 255, 0), anchor=(0, 0)))  # , html='<div style="text-align: center">'


# for i,c in enumerate():
#    p.addItem(pg.TextItem(text=c,color=(0,255,0),anchor=(1, i/2)))#, html='<div style="text-align: center">'

# self.graph.show()
# self.Add_Sidebar_Element(self.graph)
# self.inp_text_label.setText('test')

# self.network.NeuronGroups = [self.network.NeuronGroups[0]]
# self.network.SynapseGroups = [self.network.SynapseGroups[0]]


# self.network.simulate_iterations(1000, 100, measure_block_time=True)

# self.ng_width = self.network[self.group_name].width
# self.ng_height = self.network[self.group_name].height

# self.inh_ng_width = self.network[self.inh_group_name].width
# self.inh_ng_height = self.network[self.inh_group_name].height


import sys
sys.path.append('../../')

from Exploration.UI_Base import *
from functools import partial
import PymoNNto.Exploration.Evolution_Old.Computing_Devices as comp_dev
import os

class EVO_UI(UI_Base):

    def refresh_status(self, evo_server):
        status='...'
        if evo_server.Evo_Select_Combo_Box.currentText() != '':
            evo_server.running_evo_list = evo_server.running_evos()

            if evo_server.Evo_Select_Combo_Box.currentText() in evo_server.running_evo_list:
                status = '<font color="green">running</font>'
            else:
                status = '<font color="red">stopped</font>'

            evo_server.status_label.setToolTip(str(evo_server.running_evo_list))

        evo_server.status_label.setText('Status: ' + status)
        evo_server.evo_name_edit.setText(evo_server.Evo_Select_Combo_Box.currentText())

    def plot_data(self, evo_server, data):
        evo_server.curve1.setData(data)

        # smoothing
        data2 = data.copy()
        if len(data) > 1:
            for _ in range(20):
                for i in range(len(data)):
                    if i == 0:
                        data2[i] = (data2[i] + data2[i + 1] + data[i]) / 3.0
                    elif i == len(data) - 1:
                        data2[i] = (data2[i] + data2[i - 1] + data[i]) / 3.0
                    else:
                        data2[i] = (data2[i] + data2[i - 1] + data2[i + 1]) / 3.0

        evo_server.curve2.setData(data2)

    def update_evo_list(self, evo_server):
        evo_list = evo_server.evolution_list()
        evo_server.Evo_Select_Combo_Box.clear()
        evo_server.Evo_Select_Combo_Box.addItems(evo_list)
        evo_server.Evo_Select_Combo_Box.setCurrentIndex(0)
        evo_server.evo_name_edit.setText(evo_server.Evo_Select_Combo_Box.currentText())
        self.refresh_status(evo_server)

    def __init__(self):
        super().__init__(None, label='EVO manager', create_sidebar=False)
        self.reduced_layout = False
        self.evo_servers = comp_dev.get_devices()

        for evo_server in self.evo_servers:
            evo_server.main_tab = self.Next_Tab(evo_server.name, stretch=0.0)

            def group_changed(evo_server):
                self.refresh_status(evo_server)
                evo_server.curve1.clear()
                evo_server.curve2.clear()#refresh...

            evo_server.Evo_Select_Combo_Box = QComboBox()
            self.Add_element(evo_server.Evo_Select_Combo_Box, stretch=4)
            evo_server.Evo_Select_Combo_Box.currentIndexChanged.connect(partial(group_changed, evo_server))

            evo_server.status_label = QLabel('Status: ...')
            self.Add_element(evo_server.status_label, stretch=4)

            def open_terminal(evo_server):
                evo_name = evo_server.Evo_Select_Combo_Box.currentText()
                if evo_name != '':
                    evo_server.open_terminal(evo_name)
                else:
                    print('no evolution selected')

            evo_server.open_terminal_btn = QPushButton('terminal', self.main_window)
            evo_server.open_terminal_btn.clicked.connect(partial(open_terminal, evo_server))
            self.Add_element(evo_server.open_terminal_btn, stretch=1)

            def refresh(evo_server):
                evo_name = evo_server.Evo_Select_Combo_Box.currentText()
                if evo_name != '':
                    data = evo_server.plot_data(evo_name)
                    self.plot_data(evo_server, data)
                else:
                    print('no evolution selected')
                self.refresh_status(evo_server)

            evo_server.refresh_btn = QPushButton('refresh plot', self.main_window)
            evo_server.refresh_btn.clicked.connect(partial(refresh, evo_server))
            self.Add_element(evo_server.refresh_btn, stretch=1)

            def stop(evo_server):
                evo_name = evo_server.Evo_Select_Combo_Box.currentText()
                if evo_name != '':
                    evo_server.terminate_screen(evo_name)
                else:
                    print('no evolution selected')
                self.refresh_status(evo_server)

            evo_server.stop_btn = QPushButton('Stop', self.main_window)
            evo_server.stop_btn.clicked.connect(partial(stop, evo_server))
            self.Add_element(evo_server.stop_btn, stretch=1)

            def remove(evo_server):
                evo_name = evo_server.Evo_Select_Combo_Box.currentText()
                if evo_name != '':
                    evo_server.terminate_screen(evo_name)
                    evo_server.remove_evo(evo_name)
                else:
                    print('no evolution selected')
                self.update_evo_list(evo_server)
                evo_server.curve1.clear()
                evo_server.curve2.clear()


            evo_server.remove_btn = QPushButton('Remove', self.main_window)
            evo_server.remove_btn.clicked.connect(partial(remove, evo_server))
            self.Add_element(evo_server.remove_btn, stretch=1)



            self.Next_H_Block(stretch=10)
            evo_server.curve1, evo_server.curve2 = self.Add_plot_curve('plot', number_of_curves=2)

            self.Next_H_Block(stretch=0.0)

            self.Add_element(QLabel('New Evolution_Old Name:'), stretch=0)

            evo_server.evo_name_edit = QLineEdit('Evo_name')
            self.Add_element(evo_server.evo_name_edit, stretch=4)

            def add(evo_server):
                evo_name = evo_server.evo_name_edit.text()
                if evo_name != '' and evo_name != evo_server.Evo_Select_Combo_Box.currentText():
                    if evo_server.send_to_all_cb.isChecked():
                        for serv in self.evo_servers:
                            serv.transfer(evo_name)
                            self.update_evo_list(serv)
                    else:
                        evo_server.transfer(evo_name)
                        self.update_evo_list(evo_server)
                else:
                    print('please enter an evolution-name')

            evo_server.send_to_all_cb = QCheckBox('to all devices')
            self.Add_element(evo_server.send_to_all_cb, stretch=0)

            evo_server.add_btn = QPushButton('transfer to server', self.main_window)
            evo_server.add_btn.clicked.connect(partial(add, evo_server))
            self.Add_element(evo_server.add_btn, stretch=1)

            self.Next_H_Block(stretch=0.0)

            self.Add_element(QLabel('Execution file:'), stretch=0)
            evo_server.Class_Select_Combo_Box = QComboBox()

            files = []
            for r, d, f in os.walk('../../Testing'):
                rd = r.replace('../../', '').replace('\\', '.').replace('/', '.')
                for file in f:
                    if not 'Common' in rd and not 'Old' in rd and not '__pycache__' in rd:
                        files.append(rd+'.'+file.replace('.py', ''))
            print(files)

            evo_server.Class_Select_Combo_Box.addItems(files)#["Testing.SORN_Grammar.Experiment_PV_SOM", "Testing.SORN_GrammarExperiment_Hierarchical"]
            evo_server.Class_Select_Combo_Box.setCurrentIndex(0)
            self.Add_element(evo_server.Class_Select_Combo_Box, stretch=4)

            self.Add_element(QLabel('Genes:'), stretch=1)
            evo_server.evo_individuals_edit = QLineEdit('[0.95, 0.4, 0.1383, 0.1698, 0.1, 0.00015, 0.04, 0.2944, 0.0006, 0.5, 0.015, 0.2944, 0.1, 0.0001, 0.87038, 0.82099, 1.5, 0.08, 15.0]')
            self.Add_element(evo_server.evo_individuals_edit, stretch=28)

            self.Next_H_Block(stretch=0.0)

            self.Add_element(QLabel('Mutation rate:'), stretch=0)
            evo_server.mutation_edit = QLineEdit("0.05")
            self.Add_element(evo_server.mutation_edit, stretch=0)

            self.Add_element(QLabel('Ind_count:'), stretch=0)
            evo_server.ind_count_edit = QLineEdit("30")
            self.Add_element(evo_server.ind_count_edit, stretch=0)

            self.Add_element(QLabel('Threads:'), stretch=0)
            evo_server.thread_count_edit = QLineEdit("4")
            self.Add_element(evo_server.thread_count_edit, stretch=0)

            self.Add_element(QLabel('N_e:'), stretch=0)
            evo_server.N_e_edit = QLineEdit("900")
            self.Add_element(evo_server.N_e_edit, stretch=0)

            self.Add_element(QLabel('TS:'), stretch=0)
            evo_server.TS_edit = QLineEdit("[1]")
            self.Add_element(evo_server.TS_edit, stretch=0)

            evo_server.distributed_evo_cb = QCheckBox('distributed evolution')
            self.Add_element(evo_server.distributed_evo_cb, stretch=0)

            evo_server.execute_btn = QPushButton('execute', self.main_window)
            def execute(evo_server):
                evo_name = evo_server.Evo_Select_Combo_Box.currentText()
                if evo_name != '':
                    #evo_name = evo_server.evo_name_edit.text()
                    individuals = evo_server.evo_individuals_edit.text()
                    import_file = evo_server.Class_Select_Combo_Box.currentText()
                    mutation = evo_server.mutation_edit.text()
                    distributed = evo_server.distributed_evo_cb.isChecked()
                    thread_count = evo_server.thread_count_edit.text()
                    ind_count = evo_server.ind_count_edit.text()

                    N_e = evo_server.N_e_edit.text()
                    TS = evo_server.TS_edit.text()

                    evo_server.execute(evo_name, ['name='+evo_name, 'max_individual_count='+ind_count, 'thread_count='+thread_count, 'individuals='+individuals, 'import_file='+import_file, 'mutation='+str(mutation), 'distributed='+str(distributed), 'N_e='+N_e, 'TS='+TS])
                    self.update_evo_list(evo_server)
                else:
                    print('no evolution selected')

            evo_server.execute_btn.clicked.connect(partial(execute, evo_server))
            self.Add_element(evo_server.execute_btn, stretch=1)




            self.update_evo_list(evo_server)



EVO_UI().show()
